#
# Copyright (c) 2022 salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
import sys
import importlib.util
from packaging import version

if sys.version_info < (3, 8):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata

get_pkg_version = importlib_metadata.version


def is_torch_available():
    """
    Checks if pytorch is available.
    """
    if importlib.util.find_spec("torch") is not None:
        _torch_version = importlib_metadata.version("torch")
        if version.parse(_torch_version) < version.parse("1.6"):
            raise EnvironmentError(f"Torch found but with version {_torch_version}. " f"The minimum version is 1.6")
        return True
    else:
        return False


def is_tf_available():
    """
    Checks if tensorflow 2.0 is available.
    """
    candidates = (
        "tensorflow",
        "tensorflow-cpu",
        "tensorflow-gpu",
        "tf-nightly",
        "tf-nightly-cpu",
        "tf-nightly-gpu",
        "intel-tensorflow",
        "intel-tensorflow-avx512",
        "tensorflow-rocm",
        "tensorflow-macos",
    )
    _tf_version = None
    for pkg in candidates:
        try:
            _tf_version = importlib_metadata.version(pkg)
            break
        except importlib_metadata.PackageNotFoundError:
            pass
    if _tf_version is not None:
        if version.parse(_tf_version) < version.parse("2"):
            raise EnvironmentError(f"Tensorflow found but with version {_tf_version}. " f"The minimum version is 2.0")
        return True
    else:
        return False


def is_transformers_available():
    """
    Checks if the `transformers` library is installed.
    """
    if importlib.util.find_spec("transformers") is not None:
        _version = importlib_metadata.version("transformers")
        if version.parse(_version) < version.parse("4.0"):
            raise EnvironmentError(f"Transformers found but with version {_version}. " f"The minimum version is 4.0")
        return True
    else:
        return False


def is_nltk_available():
    """
        Checks if the `nltk` library is installed.
        """
    if importlib.util.find_spec("nltk") is not None:
        return True
    else:
        return False



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from .fasttext import FastText
from .semantic import Semantic
from .sequential import Sequential
from .tfidf import TfIdf
from .word2vec import Word2Vec
from logai.utils.misc import is_torch_available, \
    is_transformers_available

_MODULES = [
    "FastText",
    "Semantic",
    "Sequential",
    "TfIdf",
    "Word2Vec"
]

if is_torch_available() and is_transformers_available():
    from .forecast_nn import ForecastNN
    from .logbert import LogBERT

    _MODULES += [
        "ForecastNN",
        "LogBERT"
    ]

__all__ = _MODULES



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from datetime import datetime

import pandas as pd
from attr import dataclass
from typing import Tuple, List

from merlion.models.anomaly.dbl import DynamicBaseline, DynamicBaselineConfig

from logai.algorithms.algo_interfaces import AnomalyDetectionAlgo
from logai.config_interfaces import Config
from logai.utils import constants
from logai.utils.functions import pd_to_timeseries
from logai.algorithms.factory import factory


@dataclass
class DBLDetectorParams(Config):
    """
    Dynamic Baseline Parameters. For more details on the paramaters see 
    https://opensource.salesforce.com/Merlion/v1.3.1/merlion.models.anomaly.html#module-merlion.models.anomaly.dbl.

    :param threshold: The rule to use for thresholding anomaly scores.
    :param fixed_period: ``(t0, tf)``; Train the model on all datapoints occurring between t0 and tf (inclusive).
    :param train_window: A string representing a duration of time to serve as the scope for a
        rolling dynamic baseline model.
    :param wind_sz: The window size in minutes to bucket times of day. This parameter only applied
        if a daily trend is one of the trends used.
    :param trends: The list of trends to use. Supported trends are “daily”, “weekly” and “monthly”.
    """

    threshold: float = 0.0
    fixed_period: Tuple[str, str] = None
    train_window: str = None
    wind_sz: str = "1h"
    trends: List[str] = None
    kwargs: dict = {}


@factory.register("detection", "dbl", DBLDetectorParams)
class DBLDetector(AnomalyDetectionAlgo):
    """Dynamic baseline based time series anomaly detection. This is a wrapper class for the Dynamic Baseline
    anomaly detection model from Merlion library .
    https://opensource.salesforce.com/Merlion/v1.3.1/merlion.models.anomaly.html#module-merlion.models.anomaly.dbl
    Current implementation only supports anomaly detection on the constants.LOGLINE_COUNTS class (which maintains
    frequency counts of the log events).
    """
    def __init__(self, params: DBLDetectorParams):
        dbl_config = DynamicBaselineConfig(
            fixed_period=params.fixed_period,
            train_window=params.train_window,
            wind_sz=params.wind_sz,
            trends=params.trends,
            **params.kwargs
        )
        self.model = DynamicBaseline(dbl_config)
        self.min_ts_length = 10
        self.threshold = params.threshold

    def fit(self, log_features: pd.DataFrame):
        """
        Training method of the Dynamic Baseline model.

        :param log_features: A log feature dataframe that must only contain two columns
            ['timestamp': datetime, constants.LOGLINE_COUNTS: int].
        """
        self._is_valid_ts_df(log_features)

        time_series = pd_to_timeseries(log_features)
        self.model.train(time_series)

    def predict(self, log_features: pd.DataFrame):
        """
        Predicts anomaly scores for log_feature["timestamp", constants.LOGLINE_COUNTS].

        :param log_features: A log feature dataframe that must contain two columns
            ['timestamp': datetime, 'counts': int].
        :return: A dataframe of the predicted anomaly scores, e.g., index:log_features.index.
            value: anomaly score to indicate if anomaly or not.
        """
        self._is_valid_ts_df(log_features)

        index = log_features.index
        time_series = pd_to_timeseries(log_features)
        test_pred = self.model.get_anomaly_label(time_series)
        anom_score = test_pred.to_pd()
        anom_score["trainval"] = False
        anom_score.index = index
        return anom_score

    @staticmethod
    def _is_valid_ts_df(log_feature):
        columns = log_feature.columns.values

        for c in columns:
            if c not in [constants.LOG_TIMESTAMPS, constants.LOG_COUNTS]:
                raise ValueError(
                    "log feature dataframe must only contain two columns ['{}': datetime, '{}': int]".format(
                        constants.LOG_TIMESTAMPS, constants.LOG_COUNTS
                    )
                    + "Current columns: {}".format(columns)
                )

        if constants.LOG_TIMESTAMPS not in columns:
            raise ValueError(
                "dataframe must contain {} column".format(constants.LOG_TIMESTAMPS)
            )

        if constants.LOG_COUNTS not in columns:
            raise ValueError(
                "dataframe must contain {} column".format(constants.LOG_COUNTS)
            )

        for ts in log_feature[constants.LOG_TIMESTAMPS]:
            if not isinstance(ts, datetime):
                raise ValueError("{} must be datetime".format(constants.LOG_TIMESTAMPS))



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from enum import Enum

DIGITS_SUB = "[DIGITS]"
TIMESTAMP = "[TIMESTAMP]"


# Log record object fields
class Field(str, Enum):
    TIMESTAMP = "timestamp"
    BODY = "body"
    ATTRIBUTES = "attributes"
    RESOURCE = "resource"
    SPAN_ID = "span_id"
    LABELS = "labels"


# Attribute names
LOGLINE_NAME = "logline"
NEXT_LOGLINE_NAME = "next_logline"
PARSED_LOGLINE_NAME = "parsed_logline"
PARAMETER_LIST_NAME = "parameter_list"
LOG_EVENTS = "log_events"
LOG_TIMESTAMPS = "timestamp"
SPAN_ID = "span_id"
EVENT_INDEX = "event_index"
LABELS = "labels"

# Counts
LOG_COUNTS = "counts"


# HYPER PARAMS
MIN_TS_LENGTH = 10
COUNTER_AD_ALGO = ["ets", "dbl"]



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from logai.preprocess.openset_preprocessor import OpenSetPreprocessor
from logai.dataloader.data_model import LogRecordObject
from logai.preprocess.preprocessor import PreprocessorConfig
import pandas as pd
from logai.utils import constants


class ThunderbirdPreprocessor(OpenSetPreprocessor):
    """Custom Preprocessor for Open log dataset Thunderbird.
    """

    def __init__(self, config: PreprocessorConfig):
        super().__init__(config)

    def _get_ids(self, logrecord: LogRecordObject) -> pd.Series:
        """Get ids of loglines.

        :param logrecord: A log record object.
        :return: A pandas series containing the ids of te loglines.
        """
        return logrecord.span_id[constants.SPAN_ID]

    def _get_labels(self, logrecord: LogRecordObject):
        """Get anomaly detection labels of loglines.
        
        :param logrecord:  A log record object containing hdfs data.
        :return: The anomaly detection labels of loglines.
        """
        return logrecord.labels[constants.LABELS].apply(lambda x: int(x != "-"))



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from sklearn import metrics
import matplotlib.pyplot as plt
import seaborn as sns
from matplotlib.backends.backend_pdf import PdfPages
import numpy as np
import logging


def __plot_roc(x, y, label, y_name, x_name, fig_name):
    """Plotting roc curve.

    :param x:(np.array): array of x values.
    :param y:(np.array): array of y values.
    :param label:(np.array): array of label values.
    :param y_name:(str): y axis label.
    :param  x_name:(str): x axis label.
    :param fig_name:(str): figure name.
    """
    plt.plot(x, y, label=label)
    plt.ylabel(y_name)
    plt.xlabel(x_name)
    plt.legend(loc=4)
    plt.savefig(fig_name)

    plt.figure().clear()
    plt.close()
    plt.cla()
    plt.clf()


def __plot_scores_kde(scores_pos, scores_neg, fig_name):
    """Plotting kernel density estimation of positive and negative scores.

    :param scores_pos: (np.array): array of positive scores.
    :param scores_neg:(np.array): array of negative scores.
    :param fig_name: (str): figure name.
    """
    plt.rcParams["figure.figsize"] = [7.00, 3.50]
    plt.rcParams["figure.autolayout"] = True

    fig1 = plt.figure()
    sns.kdeplot(scores_pos, bw=0.5, color="blue")

    fig2 = plt.figure()
    sns.kdeplot(scores_neg, bw=0.5, color="red")

    pp = PdfPages(fig_name)
    fig_nums = plt.get_fignums()
    figs = [plt.figure(n) for n in fig_nums]
    for fig in figs:
        fig.savefig(pp, format="pdf")
    pp.close()

    fig1.clear()
    fig2.clear()
    plt.close(fig1)
    plt.close(fig2)
    plt.cla()


def compute_metrics(eval_metrics_per_instance_series, test_labels, test_counts=None):
    """Computing evaluation metric scores for anomaly detection.

    :param eval_metrics_per_instance_series:(dict): dict object consisting
        of eval metrics for each instance index.
    :param test_labels:(dict): gold labels for each instance index.
    :param test_counts:(dict): counts of each instance index.
    :raises: Exception: IndexError if the indices of eval_metrics_per_instance_series
        do not match with indices of test_labels.
    :return: list of tuples containing labels and scores computed for each index.
        - y: list of anomaly label for each instance.
        - loss_mean: list of mean loss (over all masked non-padded tokens) for each instance.
        - loss_max: list of max loss (over all masked non-padded tokens) for each instance.
        - loss_top6_mean: list of mean loss (averaged over top-k masked non-padded tokens) for each
        instance, k = 6(following LanoBERT paper https://arxiv.org/pdf/2111.09564.pdf).
        - scores_top6_max_prob: for each instance, we take the max prob. score obtained and average
        over the top-k masked (non-padded) token prediction, k = 6.
        - scores_top6_min_logprob: for each instance, we take the min logprob score obtained and average
        over the top-k masked (non-padded) token prediction, k = 6.
        - scores_top6_max_entropy: for each instance we take the max entropy score obtained and average
        over the top-k masked (non-padded) token prediction, k = 6.
    """
    eval_metrics_per_instance_series["indices"] = eval_metrics_per_instance_series[
        "indices"
    ].astype(int)
    y = []
    loss_mean = []
    loss_max = []
    loss_top6_mean = []
    scores_top6_max_prob = []
    scores_top6_min_logprob = []
    scores_top6_max_entropy = []
    for data in eval_metrics_per_instance_series.groupby("indices"):
        index = int(data[0])
        if index not in test_labels:
            raise Exception("Cannot find index in test ", index)
        max_losses = np.array(list(data[1]["max_loss"]))
        sum_losses = sum(list(data[1]["sum_loss"]))
        num_losses = sum(list(data[1]["num_loss"]))
        top6_losses_mean = np.mean(
            np.array(
                sorted(np.array(data[1]["top6_loss"]).flatten().tolist(), reverse=True)[
                    :6
                ]
            )
        )
        top6_max_prob = np.mean(
            1.0
            - np.array(
                sorted(np.array(data[1]["top6_max_prob"]).flatten().tolist())[:6]
            )
        )
        top6_min_logprob = np.mean(
            np.array(
                sorted(
                    np.array(data[1]["top6_min_logprob"]).flatten().tolist(),
                    reverse=True,
                )[:6]
            )
        )
        top6_max_entropy = np.mean(
            np.array(
                sorted(
                    np.array(data[1]["top6_max_entropy"]).flatten().tolist(),
                    reverse=True,
                )[:6]
            )
        )
        label = test_labels[index]
        if test_counts is not None:
            count = test_counts[index]
            y.extend([label] * count)
            if num_losses == 0:
                loss_mean.extend([0.0] * count)
            else:
                loss_mean.extend([sum_losses / num_losses] * count)
            loss_max.extend([np.max(max_losses)] * count)
            loss_top6_mean.extend([top6_losses_mean] * count)
            scores_top6_max_prob.extend([top6_max_prob] * count)
            scores_top6_min_logprob.extend([top6_min_logprob] * count)
            scores_top6_max_entropy.extend([top6_max_entropy] * count)
        else:
            y.append(label)
            if num_losses == 0:
                loss_mean.append(0.0)
            else:
                loss_mean.append(sum_losses / num_losses)
            loss_max.append(np.max(max_losses))
            loss_top6_mean.append(top6_losses_mean)
            scores_top6_max_prob.append(top6_max_prob)
            scores_top6_min_logprob.append(top6_min_logprob)
            scores_top6_max_entropy.append(top6_max_entropy)

    __compute_auc_roc(
        y,
        loss_mean,
        loss_max,
        loss_top6_mean,
        scores_top6_max_prob,
        scores_top6_min_logprob,
        scores_top6_max_entropy,
    )

    return


def __compute_auc_roc(
    y,
    loss_mean,
    loss_max,
    loss_top6_mean,
    scores_top6_max_prob,
    scores_top6_min_logprob,
    scores_top6_max_entropy,
    plot_graph=False,
    plot_histogram=False,
):
    """Computing AUROC for each of the type of metrics

    :param y: (list): list of anomaly labels for each instances.
    :param loss_mean: (list): list of mean loss (over all masked non-padded tokens) for each instance.
    :param loss_max: (list): list of max loss (over all masked non-padded tokens) for each instance.
    :param loss_top6_mean: (list): list of mean loss (averaged over top-k masked non-padded tokens)
        for each instance, k = 6 (following LanoBERT paper https://arxiv.org/pdf/2111.09564.pdf).
    :param scores_top6_max_prob: (list): for each instance, we take the max prob. score obtained
        and average over the top-k masked (non-padded) token prediction, k = 6.
    :param scores_top6_min_logprob: (list): for each instance, we take the min logprob score obtained
        and average over the top-k masked (non-padded) token prediction, k = 6.
    :param scores_top6_max_entropy: (list): for each instance we take the max entropy score obtained
        and average over the top-k masked (non-padded) token prediction, k = 6.
    :param plot_graph: (bool, optional): whether to plot roc graph. Defaults to False.
    :param plot_histogram: (bool, optional): whether to plot scores histogram. Defaults to False.
    """

    __compute_auc_roc_for_metric(
        y=y,
        metric=loss_mean,
        metric_name_str="loss_mean",
        plot_graph=plot_graph,
        plot_histogram=plot_histogram,
    )
    __compute_auc_roc_for_metric(
        y=y,
        metric=loss_max,
        metric_name_str="loss_max",
        plot_graph=plot_graph,
        plot_histogram=plot_histogram,
    )
    __compute_auc_roc_for_metric(
        y=y,
        metric=loss_top6_mean,
        metric_name_str="loss_top6_mean",
        plot_graph=plot_graph,
        plot_histogram=plot_histogram,
    )
    __compute_auc_roc_for_metric(
        y=y,
        metric=scores_top6_max_prob,
        metric_name_str="scores_top6_max_prob",
        plot_graph=plot_graph,
        plot_histogram=plot_histogram,
    )  # Note that avg scores printed for this metric would be 1 - actual probability
    __compute_auc_roc_for_metric(
        y=y,
        metric=scores_top6_min_logprob,
        metric_name_str="scores_top6_min_logprob",
        plot_graph=plot_graph,
        plot_histogram=plot_histogram,
    )
    __compute_auc_roc_for_metric(
        y=y,
        metric=scores_top6_max_entropy,
        metric_name_str="scores_top6_max_entropy",
        plot_graph=plot_graph,
        plot_histogram=plot_histogram,
    )


def __compute_auc_roc_for_metric(
    y, metric, metric_name_str, plot_graph=False, plot_histogram=False
):
    """Computing AUROC for each metric.

    :param y: (list): list of anomaly labels for each instance.
    :param metric: (list): list of metric scores for each instance.
    :param metric_name_str: (str): name of metric.
    :param plot_graph: (bool, optional): Whether to plot ROC graph. Defaults to False.
    :param plot_histogram: (bool, optional): Whether to plot histogram of metric scores. Defaults to False.
    """

    metric_pos = np.array([metric[i] for i in range(len(metric)) if y[i] == 1])
    metric_neg = np.array([metric[i] for i in range(len(metric)) if y[i] == 0])

    if metric_pos.shape[0] > 0:
        logging.info(
            "{} Pos scores:  mean: {}, std: {}".format(
                metric_name_str, np.mean(metric_pos), np.std(metric_pos)
            )
        )
    if metric_neg.shape[0] > 0:
        logging.info(
            "{} Neg scores: mean: {}, std: {}".format(
                metric_name_str, np.mean(metric_neg), np.std(metric_neg)
            )
        )

    try:
        fpr_mean, tpr_mean, thresholds = metrics.roc_curve(y, metric, pos_label=1)
        auc_mean = metrics.roc_auc_score(y, metric)
        logging.info("AUC of {}: {}".format(metric_name_str, auc_mean))

    except:
        pass

    logging.info("\n")

    if plot_graph:
        __plot_roc(
            fpr_mean,
            tpr_mean,
            "AUC=" + str(auc_mean),
            "tpr_mean",
            "fpr_mean",
            metric_name_str + "_auc.png",
        )
    if plot_histogram:
        __plot_scores_kde(metric_pos, metric_neg, metric_name_str + "_hist.pdf")



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
from attr import dataclass

from logai.algorithms.algo_interfaces import AnomalyDetectionAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


def _compute_probs(data, n=10):
    h, e = np.histogram(data, n)
    p = h / np.linalg.norm(h, ord=1)
    return e, p


def _kl_divergence(p, q):
    p[p == 0] = 1e-20
    q[q == 0] = 1e-20
    return np.sum(p * np.log(p / q))


def _js_divergence(p, q):
    m = (1.0 / 2.0) * (p + q)
    return (1.0 / 2.0) * _kl_divergence(p, m) + (1.0 / 2.0) * _kl_divergence(q, m)


@dataclass
class DistributionDivergenceParams(Config):
    """Parameters for distribution divergence based anomaly detector.

    :param n_bins: The number of bins to use to discretize the continuous distribution into a discrete distribution
    :param type: A list of types of distribution divergences. The allowed types are Kullback–Leibler ("KL"), Jensen–Shannon
        ("JS"). It also allows a comma separated list of metrics like ("KL,JS" or "JS,KL").
    """
    n_bins: int = 100
    type: list = ["KL"]  # "KL", "JS", "KL,JS"


@factory.register("detection", "distribution_divergence", DistributionDivergenceParams)
class DistributionDivergence(AnomalyDetectionAlgo):
    """Class for Distribution Divergene based Anomaly Detection. Both during training and testing, it takes log features 
    as input and construct a parametric distribution over them. For the test data, it reports the distribution divergence
    with the training data as the anomaly score.
    """
    def __init__(self, params: DistributionDivergenceParams):
        self.n_bins = params.n_bins
        self.type = params.type
        if any([x not in ["KL", "JS", "KL,JS", "JS,KL"] for x in self.type]):
            raise Exception("Type of distribution divergence allowed are: KL, JS")
        self.train_sample = None

    def fit(self, log_features: pd.DataFrame):
        """Fit method of the distribution divergence based anomaly detector. Since it is a non-parametric model, 
        there is no training required.

        :param log_features: Log features as a pandas DataFrame object.
        """
        self.train_sample = np.array(log_features)

    def predict(self, log_features: pd.DataFrame) -> list:
        """Predict method of distribution divergence based anomaly detector. It computes the distribution divergence
        between the training distribution and the test distribution provided in predict method.

        :param log_features: The test distribution as pandas DataFrame object.
        :returns: A list of scalar anomaly scores.
        """

        log_features = np.array(log_features)
        if type(self.n_bins) == int:
            _, n_bins = np.histogram(
                np.vstack([np.array(self.train_sample), log_features]), self.n_bins
            )
        else:
            n_bins = self.n_bins
        e, p = _compute_probs(self.train_sample, n=n_bins)
        _, q = _compute_probs(log_features, n=e)

        dist_divergence_scores = []
        if "KL" in self.type:
            kl = _kl_divergence(p, q)
            dist_divergence_scores.append(kl)
        if "JS" in self.type:
            js = _js_divergence(p, q)
            dist_divergence_scores.append(js)
        return dist_divergence_scores



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from attr import dataclass

import logai.algorithms.clustering_algo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class ClusteringConfig(Config):
    """Config class for Clustering algorithms.

    :param algo_name: The algorithm name.
    :param algo_params: The algorithm parameters.
    :param custom_params: Additional customized parameters.
    """
    algo_name: str = "dbscan"
    algo_params: object = None
    custom_params: object = None

    @classmethod
    def from_dict(cls, config_dict):
        config = super(ClusteringConfig, cls).from_dict(config_dict)
        config.algo_params = factory.get_config(
            "clustering", config.algo_name.lower(), config.algo_params
        )
        return config


class Clustering:
    """
    Clustering Application class defines log clustering workflow.
    It includes which algorithm to use.
    """

    def __init__(self, config: ClusteringConfig):
        self.model = factory.get_algorithm(
            "clustering", config.algo_name.lower(), config
        )

    def fit(self, log_features: pd.DataFrame):
        """Fit method of Clustering algorithm, to train on the given log features data.

        :param log_features: The training log features data.
        """
        log_features.columns = log_features.columns.astype(str)
        self.model.fit(log_features)

    def predict(self, log_features: pd.DataFrame) -> pd.Series:
        """Predict method of Clustering algorithm, to run inference on given test log features.

        :param log_features: The test log features data.
        :return: The cluster output (label).
        """
        log_features.columns = log_features.columns.astype(str)
        return self.model.predict(log_features)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from attr import dataclass
import pickle as pkl
import os

from logai.algorithms.algo_interfaces import VectorizationAlgo
from logai.config_interfaces import Config
from logai.utils.functions import pad
from logai.algorithms.factory import factory


@dataclass
class SequentialVectorizerParams(Config):
    """Config for Sequential Vectorizer which converts a sequence of loglines to sequence of log ids.

    :param sep_token: The separator token used to separate log lines in an input log sequence.
    :param model_save_dir: The path to directory where models related to sequential vectorizer would be stored.
    :param max_token_len: The maximum token length of input.
    """

    sep_token: str = None
    model_save_dir: str = None
    max_token_len: int = None


@factory.register("vectorization", "sequential", SequentialVectorizerParams)
class Sequential(VectorizationAlgo):
    """Sequential Vectorizer to convert a sequence of loglines to sequence of log ids.

    :param params: A config object for storing parameters of Sequential Vectorizer.
    """

    def __init__(self, params: SequentialVectorizerParams):
        
        self.params = params
        self.log_padding = "<pad>"
        self.log_oov = "<oov>"
        self.model_file = os.path.join(self.params.model_save_dir, "sequential.pkl")
        self.vocab = None
        if os.path.exists(self.model_file):
            self.vocab = pkl.load(open(self.model_file, "rb"))
            self.vocab_size = len(self.vocab)
            self.log_padding_id = self.vocab[self.log_padding]
            self.log_oov_id = self.vocab[self.log_oov]

    def _clean_data(self, data):
        return data.strip()

    def fit(self, loglines: pd.Series):
        """Fit method for training the sequential vectorizer.

        :param loglines: A pandas Series object containing the dataset on
            which semantic vectorizer is trained (and the vocab is built).
            Each data instance should be a logline or sequence of loglines concatenated by separator token.
        """
        model_file = None
        if self.params.model_save_dir:
            model_file = os.path.join(self.params.model_save_dir, "sequential.pkl")
        if os.path.exists(model_file):
            self.vocab = pkl.load(open(model_file, "rb"))
            self.vocab_size = len(self.vocab)
            self.log_padding_id = self.vocab[self.log_padding]
            self.log_oov_id = self.vocab[self.log_oov]
        else:
            loglines = loglines.apply(lambda x: self._clean_data(x))
            unique_loglines = set(loglines)
            unique_loglines.add(self.log_padding)
            unique_loglines.add(self.log_oov)
            self.vocab = {k: i for i, k in enumerate(list(unique_loglines))}
            self.vocab_size = len(self.vocab)
            self.log_padding_id = self.vocab[self.log_padding]
            self.log_oov_id = self.vocab[self.log_oov]
            if model_file:
                pkl.dump(self.vocab, open(model_file, "wb"))

    def transform(self, loglines: pd.Series) -> pd.Series:
        """Transform method for applying sequential vectorizer to loglines.

        :param loglines: A pandas Series containing the data to be vectorized.
            Each data instance should be a logline or sequence of loglines concatenated by separator token.
        :return: The vectorized loglines.
        """
        indices = loglines.index
        if self.params.sep_token is not None:
            loglines = loglines.apply(
                lambda x: pad(
                    [
                        self.vocab.get(self._clean_data(xi), self.log_oov_id)
                        for xi in x.split(self.params.sep_token)
                    ],
                    max_len=self.params.max_token_len,
                    padding_value=self.log_padding_id,
                )
            )
        loglines = pd.Series(loglines, indices)
        return loglines



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from logai.utils import constants
from attr import dataclass
from logai.dataloader.data_loader import FileDataLoader
from logai.dataloader.data_model import LogRecordObject
from logai.utils.dataset_utils import split_train_dev_test_for_anomaly_detection
from logai.information_extraction.log_parser import LogParser
from logai.preprocess.openset_partitioner import OpenSetPartitioner
from logai.applications.application_interfaces import WorkFlowConfig
from logai.utils.file_utils import *
from logai.preprocess.hdfs_preprocessor import HDFSPreprocessor
from logai.preprocess.bgl_preprocessor import BGLPreprocessor
from logai.preprocess.thunderbird_preprocessor import ThunderbirdPreprocessor
from logai.analysis.nn_anomaly_detector import NNAnomalyDetector
from logai.information_extraction.log_vectorizer import LogVectorizer
from .configs.schema import config_schema
import logging
from schema import SchemaError

from logai.information_extraction.feature_extractor import (
    FeatureExtractor,
    FeatureExtractorConfig,
)

logging.basicConfig(level=logging.INFO)


def validate_config_dict(workflow_config_dict):
    """Method to validate the config dict with the schema

    :param workflow_config_dict: (dict): dict containing config for anomaly detection workflow on open log datasets
    """
    try:
        config_schema.validate(workflow_config_dict)
    except SchemaError as se:
        raise se


def get_openset_ad_config(
    config_filename: str,
    anomaly_detection_type: str,
    vectorizer_type: str,
    parse_logline: bool,
    training_type: str,
):
    """Method to dynamically set some of the config parameters based on the given arguments. List of all possible supported anomaly detection types and vectorizer types configurations can be found in the config yaml file. Avoid this function if you are directly setting all config parameters manually
    
    :param config_filename: (str): Name of the config file (currently supports hdfs and bgl)
    :param anomaly_detection_type: (str): string describing the type of anomaly detection
    :param vectorizer_type: (str): string describing the type of vectorizer.
    :param parse_logline: (bool): Whether to use log parsing or not
    :param training_type: (str): Whether to use "supervised" or "unsupervised" training
    :return: OpenSetADWorkflowConfig: config object of type OpenSetADWorkflowConfig
    """
    config_path = os.path.join(
        os.path.dirname(__file__), "configs", "{}.yaml".format(config_filename)
    )

    config_parsed = read_file(config_path)
    workflow_config_dict = config_parsed["workflow_config"]

    if parse_logline is not None:
        workflow_config_dict["parse_logline"] = parse_logline

    if workflow_config_dict["training_type"] is None:
        workflow_config_dict["training_type"] = training_type

    workflow_config_dict["nn_anomaly_detection_config"] = workflow_config_dict[
        "nn_anomaly_detection_config"
    ][anomaly_detection_type]
    workflow_config_dict["log_vectorizer_config"] = workflow_config_dict[
        "log_vectorizer_config"
    ][vectorizer_type]

    if "forecast_nn" in vectorizer_type:
        if workflow_config_dict["training_type"] == "supervised":
            workflow_config_dict["log_vectorizer_config"]["algo_param"][
                "label_type"
            ] = "anomaly"
        elif workflow_config_dict["training_type"] == "unsupervised":
            workflow_config_dict["log_vectorizer_config"]["algo_param"][
                "label_type"
            ] = "next_log"
        workflow_config_dict["open_set_partitioner_config"] = workflow_config_dict[
            "open_set_partitioner_config"
        ]["forecast_nn"]
    elif "logbert" in vectorizer_type:
        workflow_config_dict["open_set_partitioner_config"] = workflow_config_dict[
            "open_set_partitioner_config"
        ]["logbert"]

    # validate_config_dict(workflow_config_dict)
    workflow_config = OpenSetADWorkflowConfig.from_dict(workflow_config_dict)
    return workflow_config


@dataclass
class OpenSetADWorkflowConfig(WorkFlowConfig):
    """Config for Log Anomaly Detection workflow on Open Log dataset Inherits: WorkFlowConfig: Config object for specifying workflow parameters

    :param dataset_name: str = None: name of the public open dataset
    :param label_filepath: str = None: path to the separate file (if any) containing the anomaly detection labels
    :param output_dir: str = None : path to output directory where all intermediate and final outputs would be dumped
    :param parse_logline: bool = False : whether to parse or not
    :param training_type: str = None: should be either supervised or unsupervised
    :param deduplicate_test: bool = False : whether to de-duplicate the instances in the test data, while maintaining a count of the number of each duplicated instance
    :param test_data_frac_pos: float = 0.8 : fraction of the logs having positive class used for teest
    :param test_data_frac_neg: float = 0.8 : fraction of the logs having negative class used for test
    :param train_test_shuffle: bool = False : whether to use chronological ordering of the logs or to shuffle them when creating the train test splits
    """

    dataset_name: str = None  # name of the public open dataset
    label_filepath: str = None  # path to the separate file (if any) containing the anomaly detection labels
    output_dir: str = None  # path to output directory where all intermediate and final outputs would be dumped
    parse_logline: bool = False  # whether to parse or not
    training_type: str = None  # should be either supervised or unsupervised
    deduplicate_test: bool = False  # whether to de-duplicate the instances in the test data, while maintaining
                                    # a count of the number of each duplicated instance
    test_data_frac_pos: float = (
        0.8  # fraction of the logs having positive class used for teest
    )
    test_data_frac_neg: float = (
        0.8  # fraction of the logs having negative class used for test
    )
    train_test_shuffle: bool = False  # whether to use chronological ordering of the logs or
                                      # to shuffle them when creating the train test splits


class OpenSetADWorkflow:
    """log anomaly detection workflow for open log datasets

    :param config: (OpenSetADWorkflowConfig): config object specifying parameters for log anomaly detection over open datasets
    """
    def __init__(self, config: OpenSetADWorkflowConfig):
        
        self.config = config

    def _get_parse_type_str(self):
        if self.config.parse_logline:
            parse_type = "parsed"
        else:
            parse_type = "nonparsed"
        return parse_type

    def _get_partition_type_str(self):
        if (
            self.config.open_set_partitioner_config.sliding_window == 0
            and self.config.open_set_partitioner_config.session_window
        ):
            partition_type = "session"
        else:
            partition_type = "sliding" + str(
                self.config.open_set_partitioner_config.sliding_window
            )
        return partition_type

    def _get_training_type_str(self):
        return self.config.training_type

    def _get_output_filename(self, suffix):
        f = pathlib.Path(self.config.data_loader_config.filepath)
        basename = f.stem + "_" + suffix
        ext = ".csv"
        filepath = basename + ext
        return os.path.join(self.config.output_dir, filepath)

    def load_dataloader(self):
        """initialize dataloader object"""
        self.dataloader = FileDataLoader(self.config.data_loader_config)

    def load_preprocessor(self):
        """initialize preprocessor object

        Raises:
            ValueError: dataset is not supported
        """
        if self.config.dataset_name == "hdfs":
            self.preprocessor = HDFSPreprocessor(
                self.config.preprocessor_config, self.config.label_filepath
            )
        elif self.config.dataset_name == "bgl":
            self.preprocessor = BGLPreprocessor(self.config.preprocessor_config)
        elif self.config.dataset_name == "thunderbird":
            self.preprocessor = ThunderbirdPreprocessor(self.config.preprocessor_config)
        else:
            raise ValueError(
                "dataset name {} not supported ".format(self.config.dataset_name)
            )

    def load_parser(self):
        """initialize log parser object"""
        self.parser = LogParser(self.config.log_parser_config)

    def load_partitioner(self):
        """initialize partitioner object"""
        self.partitioner = OpenSetPartitioner(self.config.open_set_partitioner_config)

    def load_deduper(self):
        """initialize dedup object"""
        fe_config = FeatureExtractorConfig.from_dict(
            {"group_by_category": [constants.SPAN_ID, constants.LOGLINE_NAME]}
        )
        self.feature_extractor = FeatureExtractor(fe_config)

    def load_vectorizer(self):
        """initialize vectorizer object"""
        if self.config.log_vectorizer_config.algo_param.output_dir == "":
            self.config.log_vectorizer_config.algo_param.output_dir = os.path.join(
                self.config.output_dir,
                self._get_parse_type_str()
                + "_"
                + self._get_partition_type_str()
                + "_"
                + self._get_training_type_str()
                + "_AD",
            )
        self.vectorized_data_dirpath = os.path.join(
            self.config.log_vectorizer_config.algo_param.output_dir, "vectorized_data"
        )

        if not os.path.exists(self.vectorized_data_dirpath):
            os.makedirs(self.vectorized_data_dirpath)

        self.vectorizer = LogVectorizer(self.config.log_vectorizer_config)

    def load_anomaly_detector(self):
        """initialize anomaly detector object"""
        self.set_anomaly_detector_configs()
        self.anomaly_detector = NNAnomalyDetector(
            config=self.config.nn_anomaly_detection_config
        )

    def load_data(self):
        """loads logrecord object from raw log dataset

        :return: LogRecordObject : logrecord object created from the raw log dataset
        """
        self.load_dataloader()
        logrecord = self.dataloader.load_data()
        logging.info(
            "Loaded log record object from {} ".format(
                self.config.data_loader_config.filepath
            )
        )
        return logrecord

    def preprocess_log_data(self, logrecord):
        """preprocesses logrecord object by doing custom dataset specific data cleaning and formatting

        :param logrecord: (LogRecordObject): log record object to be preprocessed
        :return:  LogRecordObject: preprocessed lgo record object using custom dataset-specific preprocessing
        """
        self.load_preprocessor()
        preprocessed_filepath = self._get_output_filename(suffix="preprocessed")
        if not file_exists(preprocessed_filepath):
            logrecord = self.preprocessor.clean_log(logrecord)
            logrecord.save_to_csv(preprocessed_filepath)
            logging.info(
                "Finished preprocessing ... saved preprocessed data in {}".format(
                    preprocessed_filepath
                )
            )
        else:
            logrecord = LogRecordObject.load_from_csv(preprocessed_filepath)
            logging.info(
                "Loaded preprocessed data from {}".format(preprocessed_filepath)
            )
        return logrecord

    def parse_log_data(self, logrecord):
        """parse logrecord object by applying standard log parsers as specified in the Config

        :param logrecord: (LogRecordObject): logrecord object to be parsed
        :return: LogRecordObject: parsed logrecord object
        """
        self.load_parser()
        parsed_filepath = self._get_output_filename(suffix=self._get_parse_type_str())
        if not file_exists(parsed_filepath):
            parsed_result = self.parser.parse(logrecord.body[constants.LOGLINE_NAME])
            logrecord.body[constants.LOGLINE_NAME] = parsed_result[
                constants.PARSED_LOGLINE_NAME
            ]
            logrecord.save_to_csv(parsed_filepath)

            logging.info(
                "Finished parsing.. saved parsed data in {}".format(parsed_filepath)
            )
        else:
            logrecord = LogRecordObject.load_from_csv(parsed_filepath)
            logging.info("Loaded past parsed data from {}".format(parsed_filepath))
        return logrecord

    def partition_log_data(self, logrecord: LogRecordObject):
        """partitioning logrecord object by applying session or sliding window based partitions

        :param logrecord: (LogRecordObject): logrecord object to be partitioned
        :return: logrecord: partitioned logrecord object
        """
        self.load_partitioner()
        output_filepath_suffix = (
            self._get_parse_type_str() + "_" + self._get_partition_type_str()
        )
        partitioned_filepath = self._get_output_filename(suffix=output_filepath_suffix)
        if not file_exists(partitioned_filepath):
            logrecord = self.partitioner.partition(logrecord)
            logrecord.save_to_csv(partitioned_filepath)
            logging.info(
                "Finished partitioning.. saved partitioned data in {}".format(
                    partitioned_filepath
                )
            )
        else:
            logrecord = LogRecordObject.load_from_csv(partitioned_filepath)
            logging.info("Loaded partitioned data from {}".format(partitioned_filepath))
        return logrecord

    def generate_train_dev_test_data(self, logrecord: LogRecordObject):
        """splitting open log datasets into train dev and test splits according to the parameters specified in the config object

        :param logrecord: (LogRecordObject): logrecord object to be split into train, dev and test
        :return: - train_data: logrecord object containing training dataset.
            - dev_data: logrecord object containing dev dataset.
            - test_data: logrecord object containing test dataset
        """
        output_filepath_suffix = (
            self._get_parse_type_str()
            + "_"
            + self._get_partition_type_str()
            + "_"
            + self._get_training_type_str()
        )
        train_filepath = self._get_output_filename(
            suffix=output_filepath_suffix + "_train"
        )
        dev_filepath = self._get_output_filename(suffix=output_filepath_suffix + "_dev")
        test_filepath = self._get_output_filename(
            suffix=output_filepath_suffix + "_test"
        )
        if not (
            file_exists(train_filepath)
            and file_exists(dev_filepath)
            and file_exists(test_filepath)
        ):
            (
                train_data,
                dev_data,
                test_data,
            ) = split_train_dev_test_for_anomaly_detection(
                logrecord,
                training_type=self.config.training_type,
                test_data_frac_neg_class=self.config.test_data_frac_neg,
                test_data_frac_pos_class=self.config.test_data_frac_pos,
                shuffle=self.config.train_test_shuffle,
            )
            train_data.save_to_csv(train_filepath)
            dev_data.save_to_csv(dev_filepath)
            test_data.save_to_csv(test_filepath)
            logging.info("Created train data .. saved in {}".format(train_filepath))
            logging.info("Created dev data .. saved in {}".format(dev_filepath))
            logging.info("Created test data .. saved in {}".format(test_filepath))
        else:
            train_data = LogRecordObject.load_from_csv(train_filepath)
            dev_data = LogRecordObject.load_from_csv(dev_filepath)
            test_data = LogRecordObject.load_from_csv(test_filepath)
            logging.info("Loaded past train data from {}".format(train_filepath))
            logging.info("Loaded past dev data from {}".format(dev_filepath))
            logging.info("Loaded past test data from {}".format(test_filepath))

        return train_data, dev_data, test_data

    def dedup_data(self, logrecord: LogRecordObject):
        """Method to run deduplication of log records, where loglines having same body and span id is collapsed into a single logline. The original occurrent count values of theseloglines is added as a pandas Series object in the 'attributes' property of the logrecord object.

        :param logrecord: (LogRecordObject): logrecord object to be deduplicated
        :return: LogRecordObject: resulting logrecord object
        """
        self.load_deduper()
        old_data_len = len(logrecord.body)
        df_new = self.feature_extractor.convert_to_counter_vector(
            log_pattern=logrecord.body[constants.LOGLINE_NAME],
            attributes=logrecord.labels.join(logrecord.span_id),
            timestamps=logrecord.timestamp[constants.LOG_TIMESTAMPS],
        )
        df_new[constants.LABELS] = df_new[constants.LOG_TIMESTAMPS].apply(
            lambda x: int(sum(x) > 0)
        )
        df_new[constants.LOG_TIMESTAMPS] = df_new[constants.LOG_TIMESTAMPS].apply(
            lambda x: x[-1]
        )
        meta_data = {
            constants.Field.BODY: [constants.LOGLINE_NAME],
            constants.Field.LABELS: [constants.LABELS],
            constants.Field.SPAN_ID: [constants.SPAN_ID],
            constants.Field.ATTRIBUTES: [constants.LOG_COUNTS],
            constants.Field.TIMESTAMP: [constants.LOG_TIMESTAMPS],
        }
        new_data_len = len(df_new)
        logrecord_new = LogRecordObject().from_dataframe(df_new, meta_data=meta_data)
        logging.info(
            "Reduced data from {} to {} by removing duplicates".format(
                old_data_len, new_data_len
            )
        )
        return logrecord_new

    def run_data_processing_workflow(self):
        """Running data processing pipeline for log anomaly detection workflow

        :return: - train_data: logrecord object containing training dataset.
            - dev_data: logrecord object containing dev dataset.
            - test_data: logrecord object containing test dataset
        """
        logrecord = self.load_data()
        logrecord = self.preprocess_log_data(logrecord=logrecord)
        if self.config.parse_logline:
            logrecord = self.parse_log_data(logrecord=logrecord)
        logrecord = self.partition_log_data(logrecord=logrecord)
        train_data, dev_data, test_data = self.generate_train_dev_test_data(logrecord)
        train_data = train_data.dropna()
        dev_data = dev_data.dropna()
        test_data = test_data.dropna()
        if (
            self.config.deduplicate_test
            and "logbert" in self.config.log_vectorizer_config.algo_name
        ):
            test_data = self.dedup_data(test_data)
        return train_data, dev_data, test_data

    def vectorizer_transform(self, logrecord: LogRecordObject, output_filename=None):
        """Applying vectorization on a logrecord object based on the kind of vectorizer specific in Config

        :param logrecord: (LogRecordObject): logrecord containing data to be vectorized
        :param output_filename: (str, optional): path to output file where the vectorized log data would be dumped. Defaults to None.
        :return: vectorized_output : vectorized data
        """
        if output_filename and os.path.exists(output_filename):
            vectorized_output = pkl.load(open(output_filename, "rb"))
        else:
            vectorized_output = self.vectorizer.transform(logrecord)
            if output_filename:
                pkl.dump(vectorized_output, open(output_filename, "wb"))
        return vectorized_output

    def run_vectorizer(self, train_logrecord, dev_logrecord, test_logrecord):
        """Wrapper method for applying vectorization on train, dev and test logrecord objects

        :param train_logrecord: (LogRecordObject): logrecord object of the training dataset
        :param dev_logrecord: (LogRecordObject): logrecord object of the dev dataset
        :param test_logrecord: (LogRecordObject): logrecord object of the test dataset
        :return: - train_data : vectorized train data.
            - dev_data: vectorized dev data.
            - test_data: vectorized test data.
        """
        self.load_vectorizer()
        self.vectorizer.fit(train_logrecord)
        vectorized_train_filename = None
        vectorized_dev_filename = None
        vectorized_test_filename = None
        if self.config.log_vectorizer_config.algo_name == "forecast_nn":
            vectorized_train_filename = os.path.join(
                self.vectorized_data_dirpath, "train.pkl"
            )
            vectorized_dev_filename = os.path.join(
                self.vectorized_data_dirpath, "dev.pkl"
            )
            vectorized_test_filename = os.path.join(
                self.vectorized_data_dirpath, "test.pkl"
            )
        train_data = self.vectorizer_transform(
            train_logrecord, output_filename=vectorized_train_filename
        )
        logging.info("Converted train data to vectors")
        dev_data = self.vectorizer_transform(
            dev_logrecord, output_filename=vectorized_dev_filename
        )
        logging.info("Converted dev data to vectors")
        test_data = self.vectorizer_transform(
            test_logrecord, output_filename=vectorized_test_filename
        )
        logging.info("Converted test data to vectors")

        return train_data, dev_data, test_data

    def run_anomaly_detection(self, train_data, dev_data, test_data):
        """Method to train and run inference of anomaly detector

        :param train_data: vectorized version of the train dataset
        :param dev_data: vectorized version of the dev dataset
        :param test_data: vectorized version of the test dataset
        """
        self.load_anomaly_detector()
        self.anomaly_detector.fit(train_data, dev_data)
        logging.info("Trained anomaly detector")
        predict_results = self.anomaly_detector.predict(test_data)
        logging.info("Ran inference on anomaly detector")

    def set_anomaly_detector_configs(self):
        """setting anomaly detector model configs based on the vectorizer configs"""
        vectorizer_params = self.config.log_vectorizer_config.algo_param
        model_params = self.config.nn_anomaly_detection_config.algo_params

        if self.config.log_vectorizer_config.algo_name == "forecast_nn":
            model_params.embedding_dim = vectorizer_params.embedding_dim
            model_params.feature_type = vectorizer_params.feature_type
            model_params.label_type = vectorizer_params.label_type
            model_params.output_dir = vectorizer_params.output_dir
            model_params.metadata_filepath = (
                vectorizer_params.vectorizer_metadata_filepath
            )

            if self.config.nn_anomaly_detection_config.algo_name == "lstm":
                model_params.max_token_len = vectorizer_params.max_token_len

        elif self.config.log_vectorizer_config.algo_name == "logbert":
            model_params.tokenizer_dirpath = vectorizer_params.tokenizer_dirpath
            model_params.model_name = vectorizer_params.model_name
            model_params.output_dir = vectorizer_params.output_dir

    def execute(self):
        """Method to execute the end to end workflow for anomaly detection on open log datasets"""
        logging.info("Going to data processing")
        (
            train_logrecord,
            dev_logrecord,
            test_logrecord,
        ) = self.run_data_processing_workflow()
        logging.info("Going to vectorize logs")
        train_data, dev_data, test_data = self.run_vectorizer(
            train_logrecord, dev_logrecord, test_logrecord
        )
        logging.info("Going to Anomaly Detection")
        self.run_anomaly_detection(train_data, dev_data, test_data)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import torch
import torch.nn.functional as F
from torch import nn
from logai.algorithms.vectorization_algo.forecast_nn import ForecastNNVectorizedDataset

from logai.algorithms.nn_model.forecast_nn.base_nn import (
    ForecastBasedNN,
    ForecastBasedNNParams,
)
from attr import dataclass


@dataclass
class CNNParams(ForecastBasedNNParams):
    """Config for CNN based log representation learning.
    
    :param kernel_sizes: the kernel size (default value: list = [2, 3, 4]).
    """

    kernel_sizes: list = [2, 3, 4]


class CNN(ForecastBasedNN):
    """
    CNN based model for learning log representation through a self-supervised forecasting task over log sequences.

    :param config: parameters for CNN log representation learning model.
    """

    def __init__(self, config: CNNParams):
        
        super().__init__(config)
        self.config = config
        self.config.model_name = "cnn"
        num_labels = self.meta_data["num_labels"]
        self.hidden_size = self.config.hidden_size

        # if isinstance(self.config.kernel_sizes, str):
        #    self.config.kernel_sizes = list(map(int, self.config.kernel_sizes.split()))

        self.convs = nn.ModuleList(
            [
                nn.Conv2d(1, self.hidden_size, (K, self.config.embedding_dim))
                for K in self.config.kernel_sizes
            ]
        )

        self.criterion = nn.CrossEntropyLoss()
        self.prediction_layer = nn.Linear(
            self.hidden_size * len(self.config.kernel_sizes), num_labels
        )

    def forward(self, input_dict):
        """Forward method for cnn model.

        :param input_dict : dict containing the session_idx, features, window_anomalies and window_labels as in ForecastNNVectorizedDataset object.
        :return: dict containing loss and prediction tensor.
        """
        if self.label_type == "anomaly":
            y = input_dict[ForecastNNVectorizedDataset.window_anomalies].long().view(-1)
        elif self.label_type == "next_log":
            y = input_dict[ForecastNNVectorizedDataset.window_labels].long().view(-1)
        self.batch_size = y.size()[0]
        x = input_dict[ForecastNNVectorizedDataset.features]
        x = self.embedder(x)

        x = x.unsqueeze(1)

        x = [
            F.relu(conv(x.float())).squeeze(3) for conv in self.convs
        ]  # [(batch_size, hidden_size, seq_len), ...]*len(kernel_sizes)
        x = [
            F.max_pool1d(i, i.size(2)).squeeze(2) for i in x
        ]  # [(batch_size, hidden_size), ...] * len(kernel_sizes)
        representation = torch.cat(x, 1)
        logits = self.prediction_layer(representation)
        y_pred = logits.softmax(dim=-1)
        loss = self.criterion(logits, y)
        return_dict = {"loss": loss, "y_pred": y_pred}
        return return_dict



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#

import pandas as pd
from attr import dataclass

from logai.algorithms.categorical_encoding_algo.label_encoding import LabelEncoding
from logai.algorithms.categorical_encoding_algo.one_hot_encoding import (
    OneHotEncoding,
    OneHotEncodingParams,
)
from logai.algorithms.categorical_encoding_algo.ordinal_encoding import (
    OrdinalEncoding,
    OrdinalEncodingParams,
)
from logai.config_interfaces import Config
from typing import Tuple


@dataclass
class CategoricalEncoderConfig(Config):
    """
    Categorical encoding configurations.
    """

    name: str = "label_encoder"
    params: object = None

    @classmethod
    def from_dict(cls, config_dict):
        config = super(CategoricalEncoderConfig, cls).from_dict(config_dict)
        if config.name.lower() == "one_hot_encoder":
            config.params = OneHotEncodingParams.from_dict(config.params)
        elif config.name.lower() == "ordinal_encoder":
            config.params = OrdinalEncodingParams.from_dict(config.params)
        return config


class CategoricalEncoder:
    """
    Implementation of the categorical encoder.
    """

    def __init__(self, config: CategoricalEncoderConfig):
        """
        Initializes a categorical encoder.

        :param config: Configuration of categorical encoders. Currently support label encoder,
            ordinal encoder, and one-hot encoder.
        """
        self.encoder = None
        if config.name.lower() == "label_encoder":
            self.encoder = LabelEncoding()

        elif config.name.lower() == "one_hot_encoder":
            self.encoder = OneHotEncoding(
                config.params if config.params else OneHotEncodingParams()
            )

        elif config.name.lower() == "ordinal_encoder":
            self.encoder = OrdinalEncoding(
                config.params if config.params else OrdinalEncodingParams()
            )
        else:
            raise RuntimeError(
                "Categorical Encoder {} is not defined".format(config.name)
            )

    def fit_transform(self, features: pd.Series) -> Tuple[pd.DataFrame, list]:
        """
        Transforms the str features into categories.

        :param features: A list of features.
        :return: A list of encoded features.
        """
        return self.encoder.fit_transform(features)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from logai.algorithms.algo_interfaces import NNAnomalyDetectionAlgo
from logai.algorithms.nn_model.logbert.configs import LogBERTConfig
from logai.algorithms.nn_model.logbert.train import LogBERTTrain
from logai.algorithms.nn_model.logbert.predict import LogBERTPredict
from logai.algorithms.factory import factory
from datasets import Dataset as HFDataset
import pandas as pd


@factory.register("detection", "logbert", LogBERTConfig)
class LogBERT(NNAnomalyDetectionAlgo):
    """Logbert model for anomaly detection of logs
    :param config: A config object for logbert model.
    """

    def __init__(self, config: LogBERTConfig):
        self.logbert_train = LogBERTTrain(config=config)
        self.logbert_predict = LogBERTPredict(config=config)

    def fit(self, train_data: HFDataset, dev_data: HFDataset):
        """Fit method for training logBERT model.
        
        :param train_data: The training dataset of type huggingface Dataset object.
        :param dev_data: The development dataset of type huggingface Dataset object.
        """
        self.logbert_train.fit(train_data, dev_data)

    def predict(self, test_data: HFDataset) -> pd.DataFrame:
        """Predict method for running inference on logBERT model.
        
        :param test_data: The test dataset of type huggingface Dataset object.
        :return: A pandas dataframe object containing the evaluation results for each type of metric.
        """
        return self.logbert_predict.predict(test_data)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from logai.algorithms.algo_interfaces import NNAnomalyDetectionAlgo
from logai.algorithms.nn_model.forecast_nn.base_nn import ForecastBasedNNParams
from logai.algorithms.vectorization_algo.forecast_nn import ForecastNNVectorizedDataset
from logai.algorithms.nn_model.forecast_nn.lstm import LSTM, LSTMParams
from logai.algorithms.nn_model.forecast_nn.cnn import CNN, CNNParams
from logai.algorithms.nn_model.forecast_nn.transformer import (
    Transformer,
    TransformerParams,
)
from logai.algorithms.factory import factory
from torch.utils.data import DataLoader


class ForcastBasedNeuralAD(NNAnomalyDetectionAlgo):
    """Forcasting based neural anomaly detection models taken from the deep-loglizer paper
    (https://arxiv.org/pdf/2107.05908.pdf).

    :param config: The parameters of general forecasting based neural anomaly detection models.
    """

    def __init__(self, config: ForecastBasedNNParams):
        
        self.model = None
        self.config = config

    def fit(
        self,
        train_data: ForecastNNVectorizedDataset,
        dev_data: ForecastNNVectorizedDataset,
    ):
        """The fit method to train forecasting based neural anomaly detection models.

        :param train_data: The training dataset of type ForecastNNVectorizedDataset
            (consisting of session_idx, features, window_anomalies and window_labels).
        :param dev_data: The development dataset of type ForecastNNVectorizedDataset
            (consisting of session_idx, features, window_anomalies and window_labels).
        """
        dataloader_train = DataLoader(
            train_data.dataset,
            batch_size=self.config.batch_size,
            shuffle=False,
            pin_memory=True,
        )
        dataloader_dev = DataLoader(
            dev_data.dataset,
            batch_size=self.config.batch_size,
            shuffle=False,
            pin_memory=True,
        )
        self.model.fit(train_loader=dataloader_train, dev_loader=dataloader_dev)

    def predict(self, test_data: ForecastNNVectorizedDataset):
        """The predict method to run inference of forecasting based neural anomaly detection model on test dataset.

        :param test_data: The test dataset of type ForecastNNVectorizedDataset
            (consisting of session_idx, features, window_anomalies and window_labels).
        :return: A dict containing overall evaluation results.
        """
        dataloader_test = DataLoader(
            test_data.dataset,
            batch_size=self.config.batch_size,
            shuffle=False,
            pin_memory=True,
        )
        result = self.model.predict(test_loader=dataloader_test)
        return result


@factory.register("detection", "lstm", LSTMParams)
class ForecastBasedLSTM(ForcastBasedNeuralAD):
    """Forecasting based lstm model for log anomaly detection.

    :param config: A config object containing parameters for LSTM based anomaly detection model.
    """

    def __init__(self, config: LSTMParams):
        super().__init__(config)
        self.config = config
        self.model = LSTM(config=self.config)


@factory.register("detection", "cnn", CNNParams)
class ForecastBasedCNN(ForcastBasedNeuralAD):
    """Forecasting based cnn model for log anomaly detection.

    :param config: A config object containing parameters for CNN based anomaly detection model.
    """

    def __init__(self, config: CNNParams):
        super().__init__(config)
        self.config = config
        self.model = CNN(config=self.config)


@factory.register("detection", "transformer", TransformerParams)
class ForecastBasedTransformer(ForcastBasedNeuralAD):
    """Forecasting based transformer model for log anomaly detection.

    :param config: A config object containing parameters for Transformer based anomaly detection model.
    """

    def __init__(self, config: TransformerParams):
        super().__init__(config)
        self.config = config
        self.model = Transformer(config=self.config)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from transformers import Trainer
from torch.utils.data import DataLoader, Dataset
from transformers.trainer_pt_utils import LabelSmoother, IterableDatasetShard
import torch


class Predictor(Trainer):
    """Custom Trainer object for running the inference of logBERT model for unsupervised anomaly detection.
    """

    def get_test_dataloader(self, test_dataset: Dataset) -> DataLoader:
        """
        Returns the test [`~torch.utils.data.DataLoader`]. Subclass and override this method if you want to inject some custom behavior.

        :param test_dataset: (`torch.utils.data.Dataset`, *optional*): The test dataset to use. If it is an `datasets.Dataset`, 
            columns not accepted by the `model.forward()` method are automatically removed. It must implement `__len__`.
        """

        if isinstance(test_dataset, torch.utils.data.IterableDataset):
            if self.args.world_size > 1:
                test_dataset = IterableDatasetShard(
                    test_dataset,
                    batch_size=self.args.eval_batch_size,
                    drop_last=self.args.dataloader_drop_last,
                    num_processes=self.args.world_size,
                    process_index=self.args.process_index,
                )
            return DataLoader(
                test_dataset,
                batch_size=self.args.eval_batch_size,
                collate_fn=self.data_collator,
                num_workers=self.args.dataloader_num_workers,
                pin_memory=self.args.dataloader_pin_memory,
            )

        test_sampler = self._get_eval_sampler(test_dataset)

        return DataLoader(
            test_dataset,
            sampler=test_sampler,
            batch_size=self.args.eval_batch_size,
            collate_fn=self.data_collator,
            drop_last=self.args.dataloader_drop_last,
            pin_memory=self.args.dataloader_pin_memory,
        )

    def compute_loss(self, model, inputs, return_outputs=False):
        """
        How the loss is computed by Trainer. By default, all models return the loss in the first element.
        Subclass and override for custom behavior.
        """

        if self.label_smoother is not None and "labels" in inputs:
            labels = inputs.pop("labels")
        else:
            labels = None
        if "indices" in inputs:
            indices = inputs.pop("indices")
        else:
            indices = None

        outputs = model(**inputs)
        # Save past state if it exists
        # TODO: this needs to be fixed and made cleaner later.
        if self.args.past_index >= 0:
            self._past = outputs[self.args.past_index]

        if labels is not None:
            loss = self.label_smoother(outputs, labels, indices)
        else:
            # We don't use .loss here since the model may return tuples instead of ModelOutput.
            loss = outputs["loss"] if isinstance(outputs, dict) else outputs[0]

        return (loss, outputs) if return_outputs else loss


class PredictionLabelSmoother(LabelSmoother):
    """
    Adds label-smoothing on a pre-computed output from a Transformers model.

    :param epsilon: (:obj:`float`, `optional`, defaults to 0.1): The label smoothing factor.
    :param ignore_index: (:obj:`int`, `optional`, defaults to -100): The index in the labels to ignore when computing the loss.
    """

    epsilon: float = 0.0
    ignore_index: int = -100
    eval_metrics_per_instance = [[], [], [], [], [], [], [], []]

    def __call__(self, model_output, labels, indices):
        logits = (
            model_output["logits"]
            if isinstance(model_output, dict)
            else model_output[0]
        )
        log_probs = -torch.nn.functional.log_softmax(logits, dim=-1)
        probs = torch.nn.functional.softmax(logits, dim=-1)
        if labels.dim() == log_probs.dim() - 1:
            labels = labels.unsqueeze(-1)

        padding_mask = labels.eq(self.ignore_index)

        # In case the ignore_index is -100, the gather will fail, so we replace labels by 0. The padding_mask
        # will ignore them in any case.
        labels.clamp_min_(0)

        nll_loss = log_probs.gather(dim=-1, index=labels)
        smoothed_loss = log_probs.sum(dim=-1, keepdim=True)

        nll_loss.masked_fill_(padding_mask, 0.0)

        smoothed_loss.masked_fill_(padding_mask, 0.0)

        probs.masked_fill_(padding_mask, 1.0)
        log_probs.masked_fill_(padding_mask, 0.0)

        predictive_prob = probs.max(dim=-1)[0]
        predictive_prob_top6 = torch.topk(predictive_prob, k=6, dim=1, largest=False)[
            0
        ].squeeze(dim=-1)
        predictive_logprob = log_probs.min(dim=-1)[0]
        predictive_logprob_top6 = torch.topk(predictive_logprob, k=6, dim=1)[0].squeeze(
            dim=-1
        )

        entropy = (probs * log_probs).sum(dim=-1)
        entropy_top6 = torch.topk(entropy, k=6, dim=1)[0].squeeze(dim=-1)

        nll_max_loss_indiv = nll_loss.max(dim=1)[0].squeeze(dim=-1)
        nll_sum_loss_indiv = nll_loss.sum(dim=1).squeeze(dim=-1)
        nll_num_loss_indiv = (~padding_mask).sum(dim=1).squeeze(dim=-1)
        nll_top6_loss_indiv = torch.topk(nll_loss, k=6, dim=1)[0].squeeze(dim=-1)

        self.eval_metrics_per_instance[0].extend(indices.cpu().data.numpy().tolist())
        self.eval_metrics_per_instance[1].extend(
            nll_max_loss_indiv.cpu().data.numpy().tolist()
        )
        self.eval_metrics_per_instance[2].extend(
            nll_sum_loss_indiv.cpu().data.numpy().tolist()
        )
        self.eval_metrics_per_instance[3].extend(
            nll_num_loss_indiv.cpu().data.numpy().tolist()
        )
        self.eval_metrics_per_instance[4].extend(
            nll_top6_loss_indiv.cpu().data.numpy().tolist()
        )
        self.eval_metrics_per_instance[5].extend(
            predictive_prob_top6.cpu().data.numpy().tolist()
        )
        self.eval_metrics_per_instance[6].extend(
            predictive_logprob_top6.cpu().data.numpy().tolist()
        )
        self.eval_metrics_per_instance[7].extend(
            entropy_top6.cpu().data.numpy().tolist()
        )

        # Take the mean over the label dimensions, then divide by the number of active elements (i.e. not-padded):
        num_active_elements = padding_mask.numel() - padding_mask.long().sum()
        nll_loss = nll_loss.sum() / num_active_elements
        smoothed_loss = smoothed_loss.sum() / (
            num_active_elements * log_probs.shape[-1]
        )
        return (1 - self.epsilon) * nll_loss + self.epsilon * smoothed_loss



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from logai.analysis.anomaly_detector import AnomalyDetector, AnomalyDetectionConfig
from logai.utils import constants
from sklearn.model_selection import train_test_split
import pandas as pd
import numpy as np


class HetAnomalyDetectionConfig(AnomalyDetectionConfig):
    """
    Heterogeneous Anomaly Detector Parameters.

    :param train_test_ratio: The ratio between test and training splits.
    """

    train_test_ratio: float = 0.3


class HetAnomalyDetector(AnomalyDetector):
    """
    Anomaly Detector Wrapper to handle heterogeneous log feature dataframe which include various attributes of log. For
    each attribute, we build its specific anomaly detector if the data satisfies the requirement.
    This current version only supports anomaly detection on the constants.LOGLINE_COUNTS field (i.e. frequency count of
    the log events).
    """

    def __init__(self, config: HetAnomalyDetectionConfig):
        self.anomaly_detector_config = AnomalyDetectionConfig(
            algo_name=config.algo_name,
            algo_params=config.algo_params,
            custom_params=config.custom_params,
        )
        self.train_test_ratio = config.train_test_ratio
        self.model_dict = {}

    def preprocess(self, counter_df: pd.DataFrame):
        """
        Splits raw log feature dataframe by unique attribute ID.

        :param counter_df: A log feature dataframe that must contain at least two columns
            ['timestamp': datetime, constants.LOGLINE_COUNTS: int].
            The rest of columns combinations are treated as log attribute ID.
        :return: The processed log feature dataframe.
        """
        ts_df = counter_df[[constants.LOG_COUNTS]]
        ts_df.index = counter_df[constants.LOG_TIMESTAMPS]
        counter_df["attribute"] = counter_df.drop(
            [constants.LOG_COUNTS, constants.LOG_TIMESTAMPS], axis=1
        ).apply(lambda x: "-".join(x.astype(str)), axis=1)
        attr_list = counter_df["attribute"].unique()
        return attr_list

    def fit_predict(self, log_feature: pd.DataFrame) -> pd.DataFrame:
        """
        Trains a model and predicts anomaly scores.

        :param log_features: A log feature dataframe that must contain at least two columns
            ['timestamp': datetime, constants.LOGLINE_COUNTS: int].
            The rest of columns combinations are treated as log attribute ID.
        :return: The predicted anomaly scores.
        """
        res = pd.DataFrame()
        attr_list = self.preprocess(log_feature)

        counter_df = log_feature
        for attr in attr_list:
            attr_df = counter_df[counter_df["attribute"] == attr]

            # if the length of sequence is short, skip the anomaly detection
            if attr_df.shape[0] < constants.MIN_TS_LENGTH:
                anom_score = np.repeat(0.0, attr_df.shape[0])
                trainval = np.repeat(None, attr_df.shape[0])
                timestamps = attr_df[[constants.LOG_TIMESTAMPS]].values.squeeze()
                tmp_dic = {
                    constants.LOG_TIMESTAMPS: timestamps,
                    "anom_score": anom_score,
                    "trainval": trainval,
                }
                tmp_df = pd.DataFrame(tmp_dic, index=attr_df.index)
                res = pd.concat([res, tmp_df])
            else:
                train, test = train_test_split(
                    attr_df[[constants.LOG_TIMESTAMPS, constants.LOG_COUNTS]],
                    shuffle=False,
                    train_size=self.train_test_ratio,
                )
                model = AnomalyDetector(self.anomaly_detector_config)
                train_scores = model.fit(train)
                test_scores = model.predict(test)
                tmp_df = pd.concat([train_scores, test_scores])
                res = pd.concat([res, tmp_df])
                self.model_dict[attr] = model
        assert (
            res.index == counter_df.index
        ).all(), "Res.index should be identical to counter_df.index"
        assert len(res) == len(
            counter_df.index
        ), "length of res should be equal to length of counter_df"
        return res



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import abc
import attr


class Config(abc.ABC):
    @classmethod
    def from_dict(cls, config_dict):
        """
        Loads a config from a config dict.

        :param config_dict: The config parameters in a dict.
        """
        if config_dict is None:
            config_dict = {}

        config = cls()
        for field in config.__dict__:
            # If there is a default and the value of the field is none we can assign a value
            if field in config_dict:
                setattr(config, field, config_dict[field])
        return config

    def as_dict(self):
        return attr.asdict(self)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd

from attr import dataclass

from logai.config_interfaces import Config
from logai.utils import constants


@dataclass
class PartitionerConfig(Config):
    """Config class for Partitioner. 

    :param group_by_category: The list of fields to group log data by .
    :param group_by_time: The string-type argument to specify grouping by time, supported types
        https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases.
    :param sliding_window: The sliding window length if partitioning loglines into sliding windows.
    :param sep_token: The separator token string to be used as delimiter, when grouping log data .
    :param exclude_last_window: A boolean (default false) whether to exclude the last window
        when doing sliding window based grouping of log data.
    :param exclude_smaller_windows: A boolean (default false) whether to exclude windows of
        length smaller than the given `sliding_window` argument.
    """
    group_by_category: list = None
    group_by_time: str = None
    sliding_window: int = 0
    sep_token: str = "[SEP]"
    exclude_last_window: bool = False
    exclude_smaller_windows: bool = False


def concat_logs(windows, tokens):
    partitioned_loglines = []
    for window in windows:
        partitioned_loglines.append(tokens.join(window))

    return partitioned_loglines


class Partitioner:
    def __init__(self, config: PartitionerConfig):
        self.config = config
        return

    def sliding_window(self, loglines: pd.Series) -> pd.Series:
        """
        Conducts sliding window log partitioning.
        
        :param loglines: The series of loglines.
        :return: The series of logline sequence after sliding window.
        """
        partitioned_loglines = self._sliding_window(loglines)
        return pd.Series(partitioned_loglines, name=loglines.name)

    def group_counter(self, logrecord_df: pd.DataFrame) -> pd.DataFrame:
        """
        Groups log records by given categories and return counter vectors.
        
        :param logrecord_df: The log record dataframe.
        :return: The log counter vector dataframe after grouping.
        """
        if not self._valid_columns(logrecord_df.columns.values):
            raise ValueError("Make sure logrecord has the columns to group by.")

        group_by_timestamps = self.config.group_by_time
        group_by_category = self.config.group_by_category

        if group_by_category is not None:
            selected_df = logrecord_df[group_by_category]
            group_bys = group_by_category

        else:
            selected_df = logrecord_df
            group_bys = logrecord_df.columns.drop(constants.LOG_TIMESTAMPS)

        if group_by_timestamps is not None:
            selected_df[constants.LOG_TIMESTAMPS] = logrecord_df[
                constants.LOG_TIMESTAMPS
            ]
            group_bys += [
                pd.Grouper(
                    key=constants.LOG_TIMESTAMPS,
                    freq=group_by_timestamps,
                    offset=0,
                    label="left",
                )
            ]

        grouped_df = (
            logrecord_df.groupby(by=group_bys, as_index=False)
            .size()
            .rename(columns={"size": constants.LOG_COUNTS})
        )

        return grouped_df

    def group_sliding_window(
        self, logrecord_df: pd.DataFrame, logline_col_name=constants.LOGLINE_NAME
    ) -> pd.DataFrame:
        """
        Groups log records by sliding window based on the sliding window length, and returns
        the resulting pandas dataFrame object.
        
        :param logrecord_df: A pandas dataFrame on which grouping is to be applied.
        :return: A pandas dataFrame after sliding window based grouping.
        """
        if not self._valid_columns(logrecord_df.columns):
            raise ValueError("Make sure logrecord has the columns to group by.")

        if logline_col_name not in logrecord_df.columns:
            raise ValueError(
                "Logline column {} cannot be found in logrecord_df.columns".format(
                    logline_col_name
                )
            )

        sliding_window = self.config.sliding_window
        group_by_timestamps = self.config.group_by_time
        group_by_category = self.config.group_by_category
        sep = self.config.sep_token

        res_sequences = []
        res_attributes = []
        partition_cols = []
        grouper = []

        if group_by_timestamps and group_by_category:
            partition_cols = [constants.LOG_TIMESTAMPS] + group_by_category
            grouper = [
                pd.Grouper(
                    key="timestamp", freq=group_by_timestamps, base=0, label="left"
                )
            ] + group_by_category

        elif group_by_timestamps:
            partition_cols = [constants.LOG_TIMESTAMPS]
            grouper = [
                pd.Grouper(
                    key="timestamp", freq=group_by_timestamps, base=0, label="left"
                )
            ]

        elif group_by_category:
            partition_cols = group_by_category
            grouper = group_by_category

        if grouper:
            # if grouping
            for group_values, data in logrecord_df.groupby(grouper):
                loglines = data[logline_col_name]
                if sliding_window >= 0:
                    partitioned_seq = self._sliding_window(loglines)
                else:
                    partitioned_seq = sep.join(loglines)
                res_sequences += partitioned_seq

                res_attributes += [group_values for _ in range(len(partitioned_seq))]
            res_df = pd.DataFrame(res_attributes, columns=partition_cols)
            res_df[logline_col_name] = res_sequences
        else:
            # if not grouping, run sliding window directly
            if sliding_window >= 0:
                res_sequences = self.sliding_window(logrecord_df[logline_col_name])
                res_df = pd.DataFrame(res_sequences)
            else:
                res_df = sep.join(logrecord_df[logline_col_name])

        return res_df

    def _valid_columns(self, columns: list):
        if self.config.group_by_time is not None:
            if constants.LOG_TIMESTAMPS not in columns:
                return False

        if self.config.group_by_category is not None:
            for col in self.config.group_by_category:
                if col not in columns:
                    return False

        return True

    def _sliding_window(self, loglines: pd.Series) -> list:
        if self.config.sliding_window <= 0:
            return list(loglines)
        if self.config.exclude_smaller_windows:
            if len(loglines) <= self.config.sliding_window:
                if self.config.exclude_last_window:
                    loglines = [self.config.sep_token.join(list(loglines)[:-1])]
                    return loglines
                else:
                    loglines = [self.config.sep_token.join(loglines)]
                    return loglines
        closed = None
        if self.config.exclude_last_window:
            closed = "left"
        windows = loglines.rolling(
            window=self.config.sliding_window,
            min_periods=self.config.sliding_window,
            closed=closed,
        )
        if self.config.exclude_smaller_windows:
            windows = list(
                filter(lambda x: len(x) >= self.config.sliding_window, windows)
            )
        windows = list(map(lambda x: self.config.sep_token.join(x), windows))
        return windows



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from .birch import BirchAlgo
from .dbscan import DbScanAlgo
from .kmeans import KMeansAlgo

__all__ = [
    "BirchAlgo",
    "DbScanAlgo",
    "KMeansAlgo",
]



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from logai.preprocess.openset_preprocessor import OpenSetPreprocessor
from logai.preprocess.preprocessor import PreprocessorConfig
import pandas as pd
from logai.utils import constants
from logai.dataloader.data_model import LogRecordObject


class HDFSPreprocessor(OpenSetPreprocessor):
    """
    Custom Preprocessor for open log dataset HDFS.
    """

    def __init__(self, config: PreprocessorConfig, label_file: str):
        super().__init__(config)
        self.id_separator = " "
        self.label_file = label_file

    def _get_labels(self, logrecord: LogRecordObject):
        """Get anomaly detection labels of loglines.
        
        :param: logrecord:  logrecord object containing hdfs data.
        :return: pd.Series object containing the anomaly detection labels of loglines.
        """
        blk_df = pd.read_csv(self.label_file, header=0)
        anomaly_blk = set(blk_df[blk_df["Label"] == "Anomaly"]["BlockId"])
        block_ids = logrecord.span_id[constants.SPAN_ID].apply(
            lambda x: set(
                self.serial_id_to_predefined_id_map[x].split(self.id_separator)
            )
        )
        labels = block_ids.apply(lambda x: int(len(x.intersection(anomaly_blk)) > 0))
        return labels

    def _get_ids(self, logrecord: LogRecordObject):
        """Get ids of loglines.
        
        :param logrecord: logrecord object containing hdfs data.
        :return:pd.Series object containing the ids of the loglines.
        """
        predefined_ids = logrecord.body[" BLOCK "]
        predefined_ids = predefined_ids.apply(lambda x: self.id_separator.join(set(x)))
        self.predefined_to_serial_id_map = {
            k: i for i, k in enumerate(list(predefined_ids))
        }
        self.serial_id_to_predefined_id_map = {
            v: k for k, v in self.predefined_to_serial_id_map.items()
        }
        predefined_ids = predefined_ids.apply(
            lambda x: self.predefined_to_serial_id_map[x]
        ).astype(int)
        return predefined_ids



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from datasets import Dataset as HFDataset
from logai.algorithms.vectorization_algo.forecast_nn import ForecastNNVectorizedDataset
from logai.analysis.anomaly_detector import AnomalyDetectionConfig
from logai.algorithms.factory import factory

NNAnomalyDetectionConfig = AnomalyDetectionConfig


class NNAnomalyDetector:
    def __init__(self, config: NNAnomalyDetectionConfig):
        """
        Initializes the anomaly detector with proper configuration. If no config provided, use default.

        :param config: A config object for anomaly detection.
        """
        self.anomaly_detector = factory.get_algorithm(
            "detection", config.algo_name.lower(), config
        )

    def fit(
        self,
        train_data: ForecastNNVectorizedDataset or HFDataset,
        dev_data: ForecastNNVectorizedDataset or HFDataset,
    ):
        """
        Trains an anomaly detection given the training and validation datasets.

        :param train_data: The training dataset.
        :param dev_data: The validation dataset
        """
        return self.anomaly_detector.fit(train_data, dev_data)

    def predict(self, test_data: ForecastNNVectorizedDataset or HFDataset):
        """
        Predicts anomalies given the test dataset.

        :param test_data: The test dataset.
        :return: A pandas dataframe containing the prediction results.
        """
        return self.anomaly_detector.predict(test_data)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import gensim
import numpy as np
import pandas as pd
import string
import pickle as pkl
import os
import gensim.downloader
import logging
from attr import dataclass
from nltk.tokenize import word_tokenize

from logai.algorithms.algo_interfaces import VectorizationAlgo
from logai.config_interfaces import Config
from logai.utils.functions import pad
from logai.algorithms.factory import factory


@dataclass
class SemanticVectorizerParams(Config):
    """
    Configuration of Semantic vectorization of loglines (or sequence of log lines) using models like word2vc, glove and fastText.

    :param max_token_len: maximum token length of the input.
    :param min_token_count: minimum count of occurrences of a token in training data for it to be considered in the vocab.
    :param sep_token: separator token used to separate log lines in input log sequence. Default is "[SEP]".
    :param embedding_dim: embedding dimension of the learnt token embeddings.
    :param window: window size parameter for word2vec and fastText models.
    :param embedding_type: type of embedding, currently supports glove, word2vec and fastText. Default is "fasttext".
    :param model_save_dir: path to directory where vectorizer models would be saved.
    """

    max_token_len: int = 10
    min_token_count: int = 1
    sep_token: str = "[SEP]"
    embedding_dim: int = 300
    window: int = 3
    embedding_type: str = "fasttext"
    model_save_dir: str = None


@factory.register("vectorization", "semantic", SemanticVectorizerParams)
class Semantic(VectorizationAlgo):
    """
    Semantic vectorizer to convert loglines into token ids based on a embedding model and vocabulary
    (like word2vec, glove and fastText). It supports either pretrained models and pretrained vocabulary
    or training word embedding models like Word2Vec or FastText on the given training data.

    :param params: A config object for semantic vectorizer.
    """

    def __init__(self, params: SemanticVectorizerParams):
        
        self.params = params
        self.model = None
        self.vocab = None
        self.vocab_size = None
        self.embed_matrix = None
        self.vocab_filename = None
        self.embed_mat_filename = None
        if self.params.model_save_dir:
            self.embed_mat_filename = os.path.join(
                self.params.model_save_dir, "embedding_matrix.npy"
            )
            self.vocab_filename = os.path.join(self.params.model_save_dir, "vocab.pkl")

        self.train_embedding_model = False

        if os.path.exists(self.vocab_filename) and os.path.exists(
            self.embed_mat_filename
        ):
            self.vocab = pkl.load(open(self.vocab_filename, "rb"))
            self.embed_matrix = np.load(self.embed_mat_filename)
            self.params.embedding_dim = self.embed_matrix.shape[1]
            self.vocab_size = len(self.vocab)

    def _tokenize_logline(self, sentence):
        try:
            sentence = sentence.translate(
                str.maketrans(string.punctuation, " " * len(string.punctuation))
            )
        except Exception as e:
            logging.info("Cannot process line: {} ".format(sentence))
            sentence = ""
        token_list = word_tokenize(sentence.lower())
        return token_list

    def fit(self, loglines: pd.Series):
        """Fit method to train semantic vectorizer.

        :param loglines: A pandas Series object containing the dataset on
            which semantic vectorizer is trained (and the vocab is built).
            Each data instance should be a logline or sequence of loglines
            concatenated by separator token.
        """
        if (
            self.params.model_save_dir
            and os.path.exists(self.vocab_filename)
            and os.path.exists(self.embed_mat_filename)
        ):
            self.vocab = pkl.load(open(self.vocab_filename, "rb"))
            self.embed_matrix = np.load(self.embed_mat_filename)
            self.params.embedding_dim = self.embed_matrix.shape[1]
            self.vocab_size = len(self.vocab)
        else:
            doc = []
            for sentence in loglines:
                token_list = self._tokenize_logline(sentence)
                doc.extend(token_list)

            doc_words = set(doc)
            if self.params.embedding_type.lower() == "glove":
                # Use Word2Vec for vectorization
                if self.train_embedding_model:
                    self.model = gensim.models.Word2Vec(
                        doc,
                        min_count=self.params.min_token_count,
                        vector_size=self.params.embedding_dim,
                        window=self.params.window,
                    )
                else:
                    if self.params.embedding_dim in [50, 100, 200, 300]:
                        self.model = gensim.downloader.load(
                            "glove-wiki-gigaword-" + str(self.embed_dim)
                        )
                    else:
                        raise ValueError(
                            "embedding dim supported for glove pretrained model\
                                 is any of (50, 100, 200, 300)"
                        )

            elif self.params.embedding_type.lower() == "word2vec":
                if self.train_embedding_model:
                    self.model = gensim.models.Word2Vec(
                        doc,
                        min_count=self.params.min_token_count,
                        vector_size=self.params.embedding_dim,
                        window=self.params.window,
                    )
                else:
                    if self.params.embedding_dim != 300:
                        raise ValueError(
                            "embedding dim supported for word2vec pretrained model is 300"
                        )
                    self.model = gensim.downloader.load("word2vec-google-news-300")

            elif self.params.embedding_type.lower() == "fasttext":
                if self.train_embedding_model:
                    self.model = gensim.models.FastText(
                        doc,
                        min_count=self.params.min_token_count,
                        vector_size=self.params.embedding_dim,
                        window=self.params.window,
                    )
                else:
                    if self.params.embedding_dim != 300:
                        raise ValueError(
                            "embedding dim supported for fasttext pretrained model is 300"
                        )
                    self.model = gensim.downloader.load(
                        "fasttext-wiki-news-subwords-300"
                    )

            if self.train_embedding_model:
                word_vectors = self.model.wv
                zero_vectors = np.zeros((3, self.params.embedding_dim))
                word_vectors.add_vectors(
                    ["UNK", "PAD", self.params.sep_token], zero_vectors
                )
                self.vocab = {k: i for i, k in enumerate(word_vectors.index_to_key)}
                self.embed_matrix = word_vectors.vectors
                self.vocab_size = len(self.vocab)
                if self.vocab_filename:
                    pkl.dump(self.vocab, open(self.vocab_filename, "wb"))
                if self.embed_mat_filename:
                    np.save(self.embed_mat_filename, self.embed_matrix)
            else:
                zero_vectors = np.zeros((3, self.params.embedding_dim))
                doc_words.update(["UNK", "PAD", self.params.sep_token])
                self.model.add_vectors(
                    ["UNK", "PAD", self.params.sep_token], zero_vectors
                )
                doc_words = doc_words.intersection(set(self.model.index_to_key))
                word_vectors_map = {k: i for i, k in enumerate(self.model.index_to_key)}
                doc_words_indices = [word_vectors_map[w] for w in doc_words]
                self.embed_matrix = self.model.vectors[doc_words_indices]
                self.vocab = {k: i for i, k in enumerate(doc_words)}
                self.vocab_size = len(self.vocab)
                if self.vocab_filename:
                    pkl.dump(self.vocab, open(self.vocab_filename, "wb"))
                if self.embed_mat_filename:
                    np.save(self.embed_mat_filename, self.embed_matrix)

    def transform(self, loglines: pd.Series) -> pd.Series:
        """Transform method to run semantic vectorizer on loglines.

        :param loglines: The pandas Series containing the data to be vectorized.
            Each data instance should be a logline or sequence of loglines concatenated by separator token.
        :return: The vectorized log data.
        """
        log_vectors = []
        count = 0
        for ll in loglines:
            token_list = self._tokenize_logline(ll)
            token_ids = [self.vocab.get(t, self.vocab["UNK"]) for t in token_list]
            log_vector = pad(
                np.array(token_ids),
                self.params.max_token_len,
                padding_value=self.vocab["PAD"],
            )

            log_vectors.append(log_vector)
            count += 1
        log_vector_series = pd.Series(log_vectors, index=loglines.index)
        logging.info("Finished converting loglines to token ids")
        return log_vector_series

    def summary(self):
        """
        Generate model summary.
        """
        return self.model.summary()



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from logai.utils.misc import is_torch_available, \
    is_transformers_available


class AlgorithmFactory:
    """
    The singleton factory class for all the supported algorithms.
    """

    _algorithms = {
        "detection": {},
        "parsing": {},
        "clustering": {},
        "vectorization": {},
    }
    _algorithms_with_torch = {
        "lstm", "cnn", "transformer",
        "logbert", "forecast_nn"
    }

    def __new__(cls):
        if not hasattr(cls, "instance"):
            cls.instance = super(AlgorithmFactory, cls).__new__(cls)
        return cls.instance

    @classmethod
    def register(cls, task, name, config_class):
        """
        Registers the algorithm name and the configuration class for an algorithm class.

        :param task: The task name, e.g., detection, clustering
        :param name: The algorithm name(s).
        :param config_class: The configuration class.
        """

        def wrap(algo_class):
            assert (
                task in cls._algorithms
            ), f"Unknown task {task}, please choose from {cls._algorithms.keys()}."
            names = [name] if isinstance(name, str) else name
            for algo_name in names:
                assert (
                    algo_name not in cls._algorithms[task]
                ), f"Algorithm {algo_name} has been already registered."
                cls._algorithms[task][algo_name] = (config_class, algo_class)
            return algo_class

        return wrap

    @classmethod
    def unregister(cls, task, name):
        """
        Unregisters a registered algorithm.

        :param task: The task name.
        :param name: The algorithm name.
        """
        return cls._algorithms[task].pop(name, None)

    def _check_algorithm(self, task, name):
        if name in self._algorithms_with_torch:
            if not is_torch_available() or not is_transformers_available():
                raise ImportError("Some deep learning packages are missing. "
                                  "Please install them via `pip install logai[deep-learning]`.")
        assert name in self._algorithms[task], \
            f"Unknown algorithm {name}, please choose from {self._algorithms[task].keys()}."

    def get_config_class(self, task, name):
        """
        Gets the corresponding configuration class given an algorithm name.

        :param task: The task name.
        :param name: The algorithm name.
        """
        self._check_algorithm(task, name)
        return self._algorithms[task][name][0]

    def get_algorithm_class(self, task, name):
        """
        Gets the corresponding algorithm class given an algorithm name.

        :param task: The task name.
        :param name: The algorithm name.
        """
        self._check_algorithm(task, name)
        return self._algorithms[task][name][1]

    def get_config(self, task, name, config_dict):
        """
        Gets a configuration instance given an algorithm name and a config dict.

        :param task: The task name.
        :param name: The algorithm name.
        :param config_dict: The config dictionary.
        """
        self._check_algorithm(task, name)
        return self._algorithms[task][name][0].from_dict(config_dict)

    def get_algorithm(self, task, name, config):
        """
        Gets a algorithm instance given an algorithm name and a config instance.

        :param task: The task name.
        :param name: The algorithm name.
        :param config: The config instance.
        """
        self._check_algorithm(task, name)
        config_class, algorithm_class = self._algorithms[task][name]
        if config and config.algo_params:
            assert isinstance(
                config.algo_params, config_class
            ), f"`config` must be an instance of {config_class.__name__}."
        algorithm = algorithm_class(
            config.algo_params if config and config.algo_params else config_class()
        )
        return algorithm


factory = AlgorithmFactory()



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import os
import pandas as pd
from attr import dataclass
from tokenizers import (
    decoders,
    models,
    normalizers,
    pre_tokenizers,
    processors,
    trainers,
    Tokenizer,
)
from transformers import BertTokenizerFast, AutoTokenizer
from datasets import Dataset as HFDataset

from logai.algorithms.algo_interfaces import VectorizationAlgo
from logai.config_interfaces import Config
from logai.dataloader.data_model import LogRecordObject
from logai.utils import constants
from logai.algorithms.factory import factory


@dataclass
class LogBERTVectorizerParams(Config):
    """Config class for logBERT Vectorizer

    :param model_name: name of the model , using HuggingFace standardized naming.
    :param use_fast: whether to use fast tokenization or not.
    :param truncation: whether to truncate the input to max_token_len.
    :param max_token_len: maximum token length of input, if truncation is set to true.
    :param max_vocab_size: maximum size  of the vocabulary.
    :param custom_tokens: list of custom tokens.
    :param train_batch_size: batch size during training the vectorizer.
    :param output_dir: path to directory where the output would be saved.
    :param tokenizer_dirpath: path to the tokenizer where the vectorizer (logbert tokenizer) would be saved.
    :param num_proc: number of processes to be used when tokenizing.

    """

    model_name: str = ""
    use_fast: bool = True
    truncation: bool = True
    max_token_len: int = 384
    max_vocab_size: int = 5000
    custom_tokens = []
    train_batch_size: int = 1000
    output_dir: str = None
    tokenizer_dirpath: str = None
    num_proc: int = 4


@factory.register("vectorization", "logbert", LogBERTVectorizerParams)
class LogBERT(VectorizationAlgo):
    """Vectorizer class for logbert.

    :param config: A config object for specifying
        parameters of log bert vectorizer.
    """

    def __init__(self, config: LogBERTVectorizerParams):
    
        self.config = config

        self.special_tokens = ["[UNK]", "[PAD]", "[CLS]", "[SEP]", "[MASK]"]

        if self.config.custom_tokens is not None:
            self.special_tokens.extend(self.config.custom_tokens)

        tokenizer_dirname = self.config.model_name + "_tokenizer"
        if self.config.tokenizer_dirpath == "" or self.config.tokenizer_dirpath is None:
            self.config.tokenizer_dirpath = os.path.join(
                self.config.output_dir, tokenizer_dirname
            )
        if not os.path.exists(self.config.tokenizer_dirpath):
            os.makedirs(self.config.tokenizer_dirpath)

        if not os.listdir(self.config.tokenizer_dirpath):
            self.tokenizer = Tokenizer(models.WordPiece(unk_token="[UNK]"))
            if "-uncased" in self.config.model_name:
                self.tokenizer.normalizer = normalizers.Sequence(
                    [
                        normalizers.NFD(),
                        normalizers.Lowercase(),
                        normalizers.StripAccents(),
                    ]
                )
            elif "-cased" in self.config.model_name:
                self.tokenizer.normalizer = normalizers.Sequence(
                    [normalizers.NFD(), normalizers.StripAccents()]
                )
            self.tokenizer.pre_tokenizer = pre_tokenizers.BertPreTokenizer()
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(
                self.config.tokenizer_dirpath, use_fast=self.config.use_fast
            )

    def fit(self, logrecord: LogRecordObject):
        """Fit method for training vectorizer for logbert.

        :param logrecord: A log record object containing the training
            dataset over which vectorizer is trained.
        """

        if os.listdir(self.config.tokenizer_dirpath):
            return

        cleaned_logrecord = self._clean_dataset(logrecord)
        dataset = self._get_hf_dataset(cleaned_logrecord)

        def batch_iterator():
            for i in range(0, len(dataset), self.config.train_batch_size):
                yield dataset[i : i + self.config.train_batch_size][
                    constants.LOGLINE_NAME
                ]

        trainer = trainers.WordPieceTrainer(
            vocab_size=self.config.max_vocab_size, special_tokens=self.special_tokens
        )

        self.tokenizer.train_from_iterator(batch_iterator(), trainer=trainer)

        cls_token_id = self.tokenizer.token_to_id("[CLS]")
        sep_token_id = self.tokenizer.token_to_id("[SEP]")

        self.tokenizer.post_processor = processors.TemplateProcessing(
            single=f"[CLS]:0 $A:0 [SEP]:0",
            pair=f"[CLS]:0 $A:0 [SEP]:0 $B:1 [SEP]:1",
            special_tokens=[
                ("[CLS]", cls_token_id),
                ("[SEP]", sep_token_id),
            ],
        )

        self.tokenizer.decoder = decoders.WordPiece(prefix="##")
        new_tokenizer = BertTokenizerFast(tokenizer_object=self.tokenizer)

        new_tokenizer.save_pretrained(self.config.tokenizer_dirpath)
        self.tokenizer = new_tokenizer

    def _tokenize_function(self, examples):
        return self.tokenizer(
            examples[constants.LOGLINE_NAME],
            truncation=self.config.truncation,
            padding="max_length",
            max_length=self.config.max_token_len,
        )

    def _clean_dataset(self, logrecord: LogRecordObject):
        special_tokens = self._get_all_special_tokens()
        loglines = logrecord.body[constants.LOGLINE_NAME]
        loglines_removed_special_tokens = loglines.apply(
            lambda x: " ".join(set(x.split(" ")) - set(special_tokens)).strip()
        )
        indices = list(logrecord.body.loc[loglines_removed_special_tokens != ""].index)

        logrecord = logrecord.select_by_index(indices, inplace=True)
        return logrecord

    def transform(self, logrecord: LogRecordObject):
        """Transform method for running vectorizer over logrecord object.

        :param logrecord: A log record object containing the dataset
            to be vectorized.
        :return: HuggingFace dataset object.
        """
        cleaned_logrecord = self._clean_dataset(logrecord)
        dataset = self._get_hf_dataset(cleaned_logrecord)
        tokenized_dataset = dataset.map(
            self._tokenize_function,
            batched=True,
            num_proc=self.config.num_proc,
            remove_columns=[constants.LOGLINE_NAME],
        )
        return tokenized_dataset

    def _get_all_special_tokens(self):
        special_tokens = self.special_tokens
        ignore_tokens = [".", "*", ":", "$", "_", "-", "/"]
        special_tokens.extend(ignore_tokens)
        return special_tokens

    def _get_hf_dataset(self, logrecord: LogRecordObject):
        if constants.LOG_COUNTS in logrecord.attributes:
            loglines_df = pd.DataFrame(
                {
                    constants.LOGLINE_NAME: logrecord.body[constants.LOGLINE_NAME],
                    constants.LABELS: logrecord.labels[constants.LABELS],
                    constants.LOG_COUNTS: logrecord.attributes[constants.LOG_COUNTS],
                }
            )
        else:
            loglines_df = pd.DataFrame(
                {
                    constants.LOGLINE_NAME: logrecord.body[constants.LOGLINE_NAME],
                    constants.LABELS: logrecord.labels[constants.LABELS],
                }
            )
        hf_data = HFDataset.from_pandas(loglines_df)
        return hf_data



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import logging
import re

import pandas as pd
from attr import dataclass

from logai.config_interfaces import Config
from logai.dataloader.data_model import LogRecordObject
from logai.utils import constants


@dataclass
class DataLoaderConfig(Config):
    """
    The configuration class of data loader.
    """

    filepath: str = ""
    log_type: str = "csv"
    dimensions: dict = dict()
    reader_args: dict = dict()
    infer_datetime: bool = False
    datetime_format: str = "%Y-%M-%dT%H:%M:%SZ"  # Default the ISO 8601 format example 2022-05-26T21:29:09+00:00
    open_dataset: str = None


class FileDataLoader:
    """
    Implementation of file data loader, reading log record objects from local files.
    """

    def __init__(self, config: DataLoaderConfig):
        """
        Initializes FileDataLoader by consuming the configuration.
        """
        self.config = config

    def load_data(self) -> LogRecordObject:
        """
        Loads log data with given configuration.
        Currently support file formats:
        - csv
        - tsv
        - other plain text format such as .log with proper parsing configurations

        :return: The logs read from log files and converted into LogRecordObject.
        """
        kwargs = self.config.reader_args
        fpath = self.config.filepath
        if self.config.log_type == "csv":
            df = pd.read_csv(fpath, **kwargs)
        elif self.config.log_type == "tsv":
            df = pd.read_table(fpath, **kwargs)
        elif self.config.log_type == "json":
            df = pd.read_json(fpath, **kwargs)
        else:
            df = self._read_logs(fpath)

        return self._create_log_record_object(df)

    def _read_logs(self, fpath):
        if "log_format" not in self.config.reader_args.keys():
            raise RuntimeError("log_format is needed to read free-form-text logs.")
        log_format = self.config.reader_args["log_format"]

        df = self._log_to_dataframe(fpath, log_format)

        return df

    def _log_to_dataframe(self, fpath, log_format):
        """
        Function to transform log file to dataframe.
        """
        headers = []
        splitters = re.split(r"(<[^<>]+>)", log_format)
        regex = ""
        for k in range(len(splitters)):
            if k % 2 == 0:
                splitter = re.sub(" +", "\\\s+", splitters[k])
                regex += splitter
            else:
                header = splitters[k].strip("<").strip(">")
                regex += "(?P<%s>.*?)" % header
                headers.append(header)
        regex = re.compile("^" + regex + "$")

        log_messages = []
        cnt = 0
        with open(fpath, "r") as fin:
            lines = fin.readlines()
            for line in lines:
                cnt += 1
                try:
                    match = regex.search(line.strip())
                    message = [match.group(header) for header in headers]
                    log_messages.append(message)
                except Exception as e:
                    logging.error("Read log file failed. Exception {}.".format(e))

        logdf = pd.DataFrame(log_messages, columns=headers, dtype=str)
        return logdf

    def _create_log_record_object(self, df: pd.DataFrame):
        dims = self.config.dimensions
        log_record = LogRecordObject()
        # Read all available log fields from config.

        if not dims:
            selected = pd.DataFrame(
                df.agg(lambda x: " ".join(x.values), axis=1).rename(
                    constants.LOGLINE_NAME
                )
            )
            setattr(log_record, "body", selected)
        else:
            for field in LogRecordObject.__dataclass_fields__:
                if field in dims.keys():
                    selected = df[list(dims[field])]
                    if field == "body":
                        if len(selected.columns) > 1:
                            selected = pd.DataFrame(
                                selected.agg(
                                    lambda x: " ".join(x.values), axis=1
                                ).rename(constants.LOGLINE_NAME)
                            )
                        else:
                            selected.columns = [constants.LOGLINE_NAME]
                    if field == "span_id":
                        if len(selected.columns) > 1:
                            raise RuntimeError("span_id should be single column")
                        selected.columns = [constants.SPAN_ID]

                    if field == "labels":
                        selected.columns = [constants.Field.LABELS]

                    if field == "timestamp":
                        if len(selected.columns) > 1:
                            selected = pd.DataFrame(
                                selected.agg(
                                    lambda x: " ".join(x.values), axis=1
                                ).rename(constants.LOG_TIMESTAMPS)
                            )
                        selected.columns = [constants.LOG_TIMESTAMPS]
                        if self.config.infer_datetime and self.config.datetime_format:
                            datetime_format = self.config.datetime_format
                            selected[constants.LOG_TIMESTAMPS] = pd.to_datetime(
                                selected[constants.LOG_TIMESTAMPS],
                                format=datetime_format,
                            )

                    setattr(log_record, field, selected)
        # log_record.__post_init__()
        return log_record


class DefaultDataLoader:
    # TODO: Placeholder to implement connector based data loader.
    def __init__(self):
        """
        Initializes default data loader.
        """
        self._logger = logging.Logger()



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from transformers import AutoModelForMaskedLM
from transformers import DataCollatorForLanguageModeling
from transformers import Trainer, TrainingArguments
from transformers import BertConfig, BertForMaskedLM
from datasets import Dataset as HFDataset
import math
import os
from .configs import LogBERTConfig
from .tokenizer_utils import get_tokenizer, get_tokenizer_vocab
import logging


class LogBERTTrain:
    """Class for training logBERT model to learn log representations"""

    def __init__(self, config: LogBERTConfig):

        self.config = config

        if not os.path.exists(config.output_dir):
            os.makedirs(config.output_dir)

        self.pretrain_from_scratch = self.config.pretrain_from_scratch
        self.mlm_probability = self.config.mlm_probability

        self.model_dirpath = os.path.join(
            self.config.output_dir, self.config.model_name
        )
        if not os.path.exists(self.model_dirpath):
            os.makedirs(self.model_dirpath)

        self.tokenizer = get_tokenizer(self.config.tokenizer_dirpath)
        self.custom_vocab = get_tokenizer_vocab(self.config.tokenizer_dirpath)

    def _initialize_trainer(self, model, train_dataset, dev_dataset):
        """initializing huggingface trainer object for logbert"""
        training_args = TrainingArguments(
            self.model_dirpath,
            evaluation_strategy=self.config.evaluation_strategy,
            num_train_epochs=self.config.num_train_epochs,
            learning_rate=self.config.learning_rate,
            logging_steps=self.config.logging_steps,
            per_device_train_batch_size=self.config.per_device_train_batch_size,
            per_device_eval_batch_size=self.config.per_device_eval_batch_size,
            weight_decay=self.config.weight_decay,
            save_steps=self.config.save_steps,
            eval_steps=self.config.eval_steps,
            resume_from_checkpoint=self.config.resume_from_checkpoint,
        )

        data_collator = DataCollatorForLanguageModeling(
            tokenizer=self.tokenizer,
            mlm_probability=self.config.mlm_probability,
            pad_to_multiple_of=self.config.max_token_len,
        )

        self.trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=train_dataset,
            eval_dataset=dev_dataset,
            data_collator=data_collator,
        )

    def get_model_checkpoint(self):
        """Get the latest dumped checkpoint from the model directory path mentioned in logBERTConfig.

        :return: path to model checkpoint (or name of model in case of a pretrained model from hugging face).
        """
        if os.path.exists(self.model_dirpath) and os.listdir(self.model_dirpath):
            checkpoint_dir = "checkpoint-" + str(
                max(
                    [
                        int(x.split("-")[1])
                        for x in os.listdir(self.model_dirpath)
                        if x.startswith("checkpoint-")
                    ]
                )
            )
            model_checkpoint = os.path.abspath(
                os.path.join(self.model_dirpath, checkpoint_dir)
            )
        else:
            model_checkpoint = self.config.model_name
        return model_checkpoint

    def fit(self, train_dataset: HFDataset, dev_dataset: HFDataset):
        """Fit method for training logbert model.

        :param train_dataset: training dataset of type huggingface Dataset object.
        :param dev_dataset: development dataset of type huggingface Dataset object.
        """
        model_checkpoint = self.get_model_checkpoint()

        if self.pretrain_from_scratch is False:
            model = AutoModelForMaskedLM.from_pretrained(model_checkpoint)
        else:
            vocab_size = len(self.custom_vocab)
            config = BertConfig(vocab_size=vocab_size)
            model = BertForMaskedLM(config)
            model.tokenizer = self.tokenizer

        self._initialize_trainer(model, train_dataset, dev_dataset)

        self.trainer.train()

    def evaluate(self):
        """Evaluate methof for evaluating logbert model on dev data using perplexity metric."""
        eval_results = self.trainer.evaluate()
        logging.info("Perplexity: {}".format(math.exp(eval_results["eval_loss"])))



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from attr import dataclass
from sklearn.svm import OneClassSVM

from logai.algorithms.algo_interfaces import AnomalyDetectionAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class OneClassSVMParams(Config):
    """Parameters for OneClass SVM based Anomaly Detector. For more explanations about the parameters see
    https://scikit-learn.org/stable/modules/generated/sklearn.svm.OneClassSVM.html.

    :param kernel: Specifies the kernel type to be used in the algorithm, i.e.,
        ``{‘linear’, ‘poly’, ‘rbf’, ‘sigmoid’, ‘precomputed’}``.
    :param degree: Degree of the polynomial kernel function (‘poly’).
    :param gamma: Kernel coefficient for ‘rbf’, ‘poly’ and ‘sigmoid’.
    :param coef0: Independent term in kernel function. It is only significant in ‘poly’ and ‘sigmoid’.
    :param tol: Tolerance for stopping criterion.
    :param nu: An upper bound on the fraction of training errors and a lower bound of the fraction of support vectors.
    :param shrinking: Whether to use the shrinking heuristic.
    :param cache_size: Specify the size of the kernel cache (in MB).
    :param verbose: Enable verbose output.
    """
    kernel: str = "linear"
    degree: int = 3
    gamma: str = "auto"
    coef0: float = 0.0
    tol: float = 1e-3
    nu: float = 0.5
    shrinking: bool = True
    cache_size: float = 200
    verbose: bool = False


@factory.register("detection", "one_class_svm", OneClassSVMParams)
class OneClassSVMDetector(AnomalyDetectionAlgo):
    def __init__(self, params: OneClassSVMParams):
        """
        OneClass SVM based Anomaly Detector. This is a wrapper class for the OneClassSVM model from scikit-learn library. For more details see 
        https://scikit-learn.org/stable/modules/generated/sklearn.svm.OneClassSVM.html.
        
        :param params: The parameters to control one class SVM models.
        """
        self.model = OneClassSVM(
            kernel=params.kernel,
            degree=params.degree,
            gamma=params.gamma,
            coef0=params.coef0,
            tol=params.tol,
            nu=params.nu,
            shrinking=params.shrinking,
            cache_size=params.cache_size,
            verbose=params.verbose,
        )

    def fit(self, log_features: pd.DataFrame):
        """
        Fit method to train the OneClassSVM on log data.

        :param log_features: Training log features as pandas DataFrame object.
        :return: The scores of the training dataset.
        """
        self.model.fit(log_features)
        train_scores = self.model.score_samples(log_features)
        train_scores = pd.DataFrame(train_scores, index=log_features.index)
        train_scores["trainval"] = True
        return train_scores

    def predict(self, log_features: pd.DataFrame) -> pd.Series:
        """
        Predict method to detect anomalies using OneClassSVM model on test log data.

        :param log_features: Test log features data as pandas DataFrame object.
        :return: A pandas dataframe of the predicted anomaly scores.
        """
        test_scores = self.model.predict(log_features)
        test_scores = pd.DataFrame(
            pd.Series(test_scores, index=log_features.index, name="anom_score")
        )
        test_scores["trainval"] = False
        return test_scores



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from attr import dataclass
from sklearn.neighbors import LocalOutlierFactor
import numpy as np

from logai.algorithms.algo_interfaces import AnomalyDetectionAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class LOFParams(Config):
    """Parameters of Locality Outlier Factors based Anomaly Detector .
    For more explanations of the parameters see https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.LocalOutlierFactor.html.

    :param n_neighbors: Number of neighbors to use by default for kneighbors queries.
    :param algorithm: Algorithm used to compute the nearest neighbors, e.g.,
        ``{‘auto’, ‘ball_tree’, ‘kd_tree’, ‘brute’}``.
    :param leaf_size: Leaf is size passed to BallTree or KDTree.
    :param metric: Metric to use for distance computation.
    :param p: Parameter for the Minkowski metric from ``sklearn.metrics.pairwise.pairwise_distances``.
    :param metric_params: Additional keyword arguments for the metric function.
    :param contamination: The amount of contamination of the data set, i.e. the proportion of outliers in the data set.
    :param novelty: By default, LocalOutlierFactor is only meant to be used for outlier detection (novelty=False).
        Set novelty to True if you want to use LocalOutlierFactor for novelty detection.
    :param n_jobs: The number of parallel jobs to run for neighbors search.
    """
    n_neighbors: int = 20
    algorithm: str = "auto"
    leaf_size: int = 30
    metric: callable or str = "minkowski"
    p: int = 2
    metric_params: dict = None
    contamination: str = "auto"
    novelty: bool = True
    n_jobs: int = None


@factory.register("detection", "lof", LOFParams)
class LOFDetector(AnomalyDetectionAlgo):
    """Locality Outlier Factor based Anomaly Detector. This is a wrapper method for the LOF based Detector in scikit-learn library.
    See https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.LocalOutlierFactor.html for more details.
    """
    def __init__(self, params: LOFParams):
        self.model = LocalOutlierFactor(
            n_neighbors=params.n_neighbors,
            algorithm=params.algorithm,
            leaf_size=params.leaf_size,
            metric=params.metric,
            p=params.p,
            metric_params=params.metric_params,
            contamination=params.contamination,
            novelty=params.novelty,
            n_jobs=params.n_jobs,
        )

    def fit(self, log_features: pd.DataFrame):
        """
        Fits a LOF model.

        :param log_features: The input for model training.
        :return:  pandas.Dataframe : The scores of the training dataset.
        """
        self.model.fit(
            np.array(log_features)
        )  # casting to numpy array to avoid warning on feature names
        train_scores = self.model.score_samples(log_features)
        train_scores = pd.DataFrame(train_scores, index=log_features.index)
        train_scores["trainval"] = True
        return train_scores

    def predict(self, log_features: pd.DataFrame) -> pd.Series:
        """
        Predicts anomaly scores.

        :param log_features: The input for inference.
        :return: A pandas dataframe of the predicted anomaly scores.
        """
        test_scores = self.model.predict(
            np.array(log_features)
        )  # casting to numpy array to avoid warning on feature names
        test_scores = pd.DataFrame(
            pd.Series(test_scores, index=log_features.index, name="anom_score")
        )

        test_scores["trainval"] = False
        return test_scores



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd

from logai.dataloader.data_model import LogRecordObject
from logai.utils import constants
from sklearn.model_selection import train_test_split


def split_train_dev_test_for_anomaly_detection(
        logrecord,
        training_type,
        test_data_frac_neg_class,
        test_data_frac_pos_class=None,
        shuffle=False,
):
    """Util method to split a logrecord object into train dev and test splits, where the splitting fractions 
    are based on the SPAN_ID field of the logrecord.

    :param logrecord: (LogRecordObject): input logrecord object to be split into train, dev and test splits
    :param training_type: (str): 'supervised' or 'unsupervised'
    :param test_data_frac_neg_class: (float): fraction of the negative class to be . Defaults to None.
    :param test_data_frac_pos_class: (float, optional): when supervised mode is selected, fraction of the
        positive class data to be used for test data. (fraction for dev data is fixed).For unsupervised mode this value is fixed to 1.0
    :param shuffle: (bool, optional): whether to shuffle the log data when splitting into train and test. 
        If False, then it uses the chronological ordering, where the first (chronologically first) split will
        constitute train data, second one development data and third one as test data. Defaults to False.

    :return: - logrecord_train: logrecord object containing train data.
        - logrecord_dev: logrecord object containing dev data.
        - logrecord_test: logrecord object containing test data.
    
    """

    if training_type not in ["supervised", "unsupervised"]:
        raise ValueError("training_type must be either supervised or unsupervised")

    if training_type == "unsupervised":
        test_data_frac_pos_class = 1.0
    elif test_data_frac_pos_class is None:
        test_data_frac_pos_class = test_data_frac_neg_class

    if not shuffle:
        stratify = None
    else:
        stratify = logrecord.labels[constants.LABELS]

    train_ids, test_ids, train_labels, test_labels = train_test_split(
        logrecord.span_id[constants.SPAN_ID],
        logrecord.labels[constants.LABELS],
        test_size=0.2,
        shuffle=shuffle,
        stratify=stratify,
    )
    if not shuffle:
        stratify = None
    else:
        stratify = train_labels
    train_ids, dev_ids, train_labels, dev_labels = train_test_split(
        train_ids, train_labels, test_size=0.1, shuffle=shuffle, stratify=stratify
    )

    test_ids = list(set(test_ids))

    if training_type == "unsupervised":
        train_ids_labels = pd.DataFrame({"ids": train_ids, "labels": train_labels})
        train_ids_pos = train_ids_labels[train_ids_labels["labels"] == 1]["ids"]
        dev_ids_labels = pd.DataFrame({"ids": dev_ids, "labels": dev_labels})
        dev_ids_pos = dev_ids_labels[dev_ids_labels["labels"] == 1]["ids"]

        train_ids = train_ids_labels[train_ids_labels["labels"] == 0]["ids"]
        dev_ids = dev_ids_labels[dev_ids_labels["labels"] == 0]["ids"]

        test_ids.extend(train_ids_pos)
        test_ids.extend(dev_ids_pos)

        if len(dev_ids) == 0:
            train_ids, dev_ids = train_test_split(
                train_ids, test_size=0.1, shuffle=shuffle
            )

    else:
        train_ids = list(set(train_ids))
        dev_ids = list(set(dev_ids))

    indices_train = list(
        logrecord.span_id.loc[
            logrecord.span_id[constants.SPAN_ID].isin(train_ids)
        ].index
    )
    indices_dev = list(
        logrecord.span_id.loc[logrecord.span_id[constants.SPAN_ID].isin(dev_ids)].index
    )
    indices_test = list(
        logrecord.span_id.loc[logrecord.span_id[constants.SPAN_ID].isin(test_ids)].index
    )

    print(
        "indices_train/dev/test: ",
        len(indices_train),
        len(indices_dev),
        len(indices_test),
    )
    logrecord_train = logrecord.select_by_index(indices_train)
    logrecord_dev = logrecord.select_by_index(indices_dev)
    logrecord_test = logrecord.select_by_index(indices_test)

    return logrecord_train, logrecord_dev, logrecord_test



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from attr import dataclass

from logai.analysis.anomaly_detector import AnomalyDetectionConfig
from logai.analysis.nn_anomaly_detector import NNAnomalyDetectionConfig
from logai.analysis.clustering import ClusteringConfig
from logai.config_interfaces import Config
from logai.dataloader.data_loader import DataLoaderConfig
from logai.dataloader.openset_data_loader import OpenSetDataLoaderConfig
from logai.information_extraction.categorical_encoder import CategoricalEncoderConfig
from logai.information_extraction.feature_extractor import FeatureExtractorConfig
from logai.information_extraction.log_parser import LogParserConfig
from logai.information_extraction.log_vectorizer import VectorizerConfig
from logai.preprocess.partitioner import PartitionerConfig
from logai.preprocess.openset_partitioner import OpenSetPartitionerConfig
from logai.preprocess.preprocessor import PreprocessorConfig


@dataclass
class WorkFlowConfig(Config):
    """config class for end to end workflow.
    
    :param data_loader_config: A config object for data loader.
    :param open_set_data_loader_config: A config object for data loader for opensource public log datasets.
    :param preprocessor_config: A config object for log preprocessor.
    :param log_parser_config: A config object for log parser.
    :param log_vectorizer_config: A config object for log vectorizer.
    :param partitioner_config: A config object for log partitioner.
    :param open_set_partitioner_config: A config object for log partitioner for opensource public log datasets.
    :param categorical_encoder_config: A config object for categorical encoder of log data.
    :param feature_extractor_config: A config object for log feature extractor.
    :param anomaly_detection_config: A config object for log anomaly detector.
    :param nn_anomaly_detection_config: A config object for neural anomaly detector.
    :param clustering_config: A config object for log clustering algorithm.
    """
    data_loader_config: object = None
    open_set_data_loader_config: object = None
    preprocessor_config: object = None

    log_parser_config: object = None
    log_vectorizer_config: object = None
    partitioner_config: object = None
    open_set_partitioner_config: object = None
    categorical_encoder_config: object = None
    feature_extractor_config: object = None
    anomaly_detection_config: object = None
    nn_anomaly_detection_config: object = None
    clustering_config: object = None
    workflow_config: object = None

    @classmethod
    def from_dict(cls, config_dict):
        config = super(WorkFlowConfig, cls).from_dict(config_dict)

        if config.data_loader_config:
            config.data_loader_config = DataLoaderConfig.from_dict(
                config.data_loader_config
            )

        if config.open_set_data_loader_config:
            config.open_set_data_loader_config = OpenSetDataLoaderConfig.from_dict(
                config.open_set_data_loader_config
            )

        if config.preprocessor_config:
            config.preprocessor_config = PreprocessorConfig.from_dict(
                config.preprocessor_config
            )

        if config.partitioner_config:
            config.partitioner_config = PartitionerConfig.from_dict(
                config.partitioner_config
            )

        if config.open_set_partitioner_config:
            config.open_set_partitioner_config = OpenSetPartitionerConfig.from_dict(
                config.open_set_partitioner_config
            )

        if config.log_parser_config:
            config.log_parser_config = LogParserConfig.from_dict(
                config.log_parser_config
            )

        if config.log_vectorizer_config:
            config.log_vectorizer_config = VectorizerConfig.from_dict(
                config.log_vectorizer_config
            )

        if config.feature_extractor_config:
            config.feature_extractor_config = FeatureExtractorConfig.from_dict(
                config.feature_extractor_config
            )

        if config.categorical_encoder_config:
            config.categorical_encoder_config = CategoricalEncoderConfig.from_dict(
                config.categorical_encoder_config
            )

        if config.anomaly_detection_config:
            config.anomaly_detection_config = AnomalyDetectionConfig.from_dict(
                config.anomaly_detection_config
            )

        if config.nn_anomaly_detection_config:
            config.nn_anomaly_detection_config = NNAnomalyDetectionConfig.from_dict(
                config.nn_anomaly_detection_config
            )

        if config.clustering_config:
            config.clustering_config = ClusteringConfig.from_dict(
                config.clustering_config
            )

        return config



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import torch
import os
import numpy as np
import pickle as pkl
import pandas as pd
from attr import dataclass

from logai.algorithms.algo_interfaces import VectorizationAlgo
from .semantic import Semantic, SemanticVectorizerParams
from .sequential import Sequential, SequentialVectorizerParams
from logai.config_interfaces import Config
from logai.dataloader.data_model import LogRecordObject
from logai.utils import constants
from logai.algorithms.factory import factory


class ForecastNNVectorizedDataset:
    """Class for storing vectorized dataset for forecasting based neural models.
    :param logline_features: (np.array): list of vectorized log-sequences
    :param labels: (list or pd.Series or np.array): list of labels (anomalous or non-anomalous) for each log sequence.
    :param nextlogline_ids: (list or pd.Series or np.array): list of ids of next loglines, for each log sequence
    :param span_ids: (list or pd.Series or np.array): list of ids of log sequences.
    """

    session_idx: str = "session_idx"
    features: str = "features"
    window_anomalies: str = "window_anomalies"
    window_labels: str = "window_labels"

    def __init__(self, logline_features, labels, nextlogline_ids, span_ids):
        
        self.dataset = []
        for data_i, label_i, next_i, index_i in zip(
            logline_features, labels, nextlogline_ids, span_ids
        ):
            self.dataset.append(
                {
                    self.session_idx: np.array([index_i]),
                    self.features: np.array(data_i),
                    self.window_anomalies: label_i,
                    self.window_labels: next_i,
                }
            )


@dataclass
class ForecastNNVectorizerParams(Config):
    """Config class for vectorizer for forecast based neural models for log representation learning.

    :param feature_type: The type of log feature representation where the supported types "semantics" and "sequential".
    :param label_type: The type of label, anomaly or next_log, which corresponds to the supervised and the
        forecasting based unsupervised setting.
    :param sep_token: The separator token used when constructing the log sequences during log grouping/partitioning. (default = "[SEP]")
    :param max_token_len: The maximum token length of the input.
    :param min_token_count: The minimum number of occurrences of a token in the training data, for it to be
        considered in the vocab.
    :param embedding_dim: The embedding dimension of the tokens.
    :param output_dir: The path to output directory where the vectorizer model directory and metadata file
        would be created.
    :param vectorizer_metadata_filepath: The path to file where the vectorizer metadata would be saved.
        This would be read by the anomaly detection model and should be set in the metadata_filepath of
        the forecast_nn based anomaly detector.
    :param vectorizer_model_dirpath: The path to directory containing the vectorizer model.
    """

    feature_type: str = None  # supported types "semantics" and "sequential"
    label_type: str = None
    sep_token: str = "[SEP]"
    max_token_len: int = None
    min_token_count: int = None
    embedding_dim: int = None
    output_dir: str = ""
    vectorizer_metadata_filepath: str = ""
    vectorizer_model_dirpath: str = ""

    sequentialvec_config: object = None
    semanticvec_config: object = None


@factory.register("vectorization", "forecast_nn", ForecastNNVectorizerParams)
class ForecastNN(VectorizationAlgo):
    """Vectorizer Class for forecast based neural models for log representation learning.
    
    :param config: config object specifying parameters of forecast based neural log repersentation learning model.
    """

    def __init__(self, config: ForecastNNVectorizerParams):
        self.meta_data = {}
        self.config = config
        self.config.vectorizer_model_dirpath = os.path.join(
            self.config.output_dir, "embedding_model"
        )
        self.config.vectorizer_metadata_filepath = os.path.join(
            self.config.vectorizer_model_dirpath, "metadata.pkl"
        )

        if not os.path.exists(self.config.vectorizer_model_dirpath):
            os.makedirs(self.config.vectorizer_model_dirpath)

        sequentialvec_config = SequentialVectorizerParams.from_dict(
            {
                "sep_token": self.config.sep_token,
                "max_token_len": self.config.max_token_len,
                "model_save_dir": self.config.vectorizer_model_dirpath,
            }
        )
        self.sequential_vectorizer = Sequential(sequentialvec_config)
        self.semantic_vectorizer = None
        if self.config.feature_type == "semantics":
            semanticvec_config_dict = {
                "max_token_len": self.config.max_token_len,
                "min_token_count": self.config.min_token_count,
                "sep_token": self.config.sep_token,
                "embedding_dim": self.config.embedding_dim,
                "model_save_dir": self.config.vectorizer_model_dirpath,
            }
            semanticvec_config = SemanticVectorizerParams.from_dict(
                semanticvec_config_dict
            )
            self.semantic_vectorizer = Semantic(semanticvec_config)

    def _process_logsequence(self, data_sequence):
        data_sequence = data_sequence.dropna()
        unique_data_instances = pd.Series(
            list(
                set(
                    self.config.sep_token.join(list(data_sequence)).split(
                        self.config.sep_token
                    )
                )
            )
        )
        return unique_data_instances

    def fit(self, logrecord: LogRecordObject):
        """Fit method to train vectorizer.

        :param logrecord: A log record object to train the vectorizer on.
        """
        if self.sequential_vectorizer.vocab is None or (
            self.config.feature_type == "semantics"
            and self.semantic_vectorizer.vocab is None
        ):
            loglines = logrecord.body[
                constants.LOGLINE_NAME
            ]  # data[self.config.loglines_field]
            nextloglines = logrecord.attributes[
                constants.NEXT_LOGLINE_NAME
            ]  # data[self.config.nextlog_field]
            loglines = pd.concat([loglines, nextloglines])
            loglines = self._process_logsequence(loglines)
        if self.sequential_vectorizer.vocab is None:
            self.sequential_vectorizer.fit(loglines)
        if (
            self.config.feature_type == "semantics"
            and self.semantic_vectorizer.vocab is None
        ):
            self.semantic_vectorizer.fit(loglines)
        self._dump_meta_data()

    def transform(self, logrecord: LogRecordObject):
        """Transform method to run vectorizer on logrecord object.

        :param logrecord: A log record object to be vectorized.
        :return: ForecastNNVectorizedDataset object containing the vectorized dataset.
        """
        if self.config.feature_type == "sequential":
            logline_features = self.sequential_vectorizer.transform(
                logrecord.body[constants.LOGLINE_NAME]
            )
        elif self.config.feature_type == "semantics":
            logline_features = self.semantic_vectorizer.transform(
                logrecord.body[constants.LOGLINE_NAME]
            )
        if constants.NEXT_LOGLINE_NAME in logrecord.attributes:
            nextlogline_ids = self.sequential_vectorizer.transform(
                logrecord.attributes[constants.NEXT_LOGLINE_NAME]
            ).apply(lambda x: x[0])
        else:
            nextlogline_ids = None
        labels = logrecord.labels[constants.LABELS]
        span_ids = logrecord.span_id[constants.SPAN_ID]
        samples = ForecastNNVectorizedDataset(
            logline_features=logline_features,
            labels=labels,
            nextlogline_ids=nextlogline_ids,
            span_ids=span_ids,
        )
        return samples

    def _dump_meta_data(self):
        if not os.path.exists(self.config.vectorizer_metadata_filepath):
            if self.config.feature_type == "sequential":
                self.meta_data["vocab_size"] = self.sequential_vectorizer.vocab_size
            else:
                self.meta_data["vocab_size"] = self.semantic_vectorizer.vocab_size
            if self.config.feature_type == "semantics":
                self.meta_data["pretrain_matrix"] = torch.Tensor(
                    self.semantic_vectorizer.embed_matrix
                )
            if self.config.label_type == "anomaly":
                self.meta_data["num_labels"] = 2
            else:
                self.meta_data["num_labels"] = self.sequential_vectorizer.vocab_size
            pkl.dump(
                self.meta_data, open(self.config.vectorizer_metadata_filepath, "wb")
            )
        else:
            self.meta_data = pkl.load(
                open(self.config.vectorizer_metadata_filepath, "rb")
            )



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from attr import dataclass
from logai.config_interfaces import Config


@dataclass
class LogBERTConfig(Config):
    """Config for logBERT model.
    
    :param pretrain_from_scratch: bool = True  : whether to do pretraining from scratch or intialize with the HuggingFace pretrained LM.
    :param model_name: str = "bert-base-cased" : name of the model using HuggingFace standardized naming.
    :param model_dirname: str = None : name of the directory where the model would be saved. Directory of this
        name would be created inside `output_dir`, if it does not exist.
    :param mlm_probability: float = 0.15 : probability of the tokens to be masked during MLM trainning.
    :param mask_ngram: int = 1 : length of ngrams that are masked during inference.
    :param max_token_len: int = 384 : maximum token length of the input.
    :param learning_rate: float = 1e-5 : learning rate.
    :param weight_decay: float = 0.0001 : parameter to use weight decay of the learning rate.
    :param per_device_train_batch_size: int = 50 : training batch size per gpu device.
    :param per_device_eval_batch_size: int = 256 : evaluation batch size per gpu device.
    :param eval_accumulation_steps: int = 1000 : parameter to accumulate the evaluation results over the steps.
    :param num_eval_shards: int = 10 : parameter to shard the evaluation data (to avoid any OOM issue).
    :param evaluation_strategy: str = "steps" : either steps or epoch, based on whether the unit of the eval_steps
         parameter is "steps" or "epoch".
    :param num_train_epochs: int = 20 : number of training epochs.
    :param logging_steps: int = 10 : number of steps after which the output is logged.
    :param save_steps: int = 50 : number of steps after which the model is saved.
    :param eval_steps: int = 50 : number of steps after which evaluation is run.
    :param resume_from_checkpoint: bool = True : whether to resume from a given model checkpoint. 
        If set to true, it will find the latest checkpoint saved in the dir and use that to load the model.
    :param output_dir: str = None : output directory where the model would be saved.
    :param tokenizer_dirpath: str = None : path to directory containing the tokenizer.
    """

    pretrain_from_scratch: bool = True
    model_name: str = "bert-base-cased"
    model_dirname: str = None
    mlm_probability: float = 0.15
    mask_ngram: int = 1
    max_token_len: int = 384
    evaluation_strategy: str = "steps"
    num_train_epochs: int = 20
    learning_rate: float = 1e-5
    logging_steps: int = 10
    per_device_train_batch_size: int = 50
    per_device_eval_batch_size: int = 256
    eval_accumulation_steps: int = 1000
    num_eval_shards: int = 10
    weight_decay: float = 0.0001
    save_steps: int = 50
    eval_steps: int = 50
    resume_from_checkpoint: bool = True
    output_dir: str = None
    tokenizer_dirpath: str = None



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import re

import pandas as pd


def generate_logformat_regex(log_format):
    """
    Function to generate regular expression to split log messages.

    return: headers, regex.
    """
    headers = []
    splitters = re.split(r"(<[^<>]+>)", log_format)
    regex = ""
    for k in range(len(splitters)):
        if k % 2 == 0:
            splitter = re.sub(" +", "\\\s+", splitters[k])
            regex += splitter
        else:
            header = splitters[k].strip("<").strip(">")
            regex += "(?P<%s>.*?)" % header
            headers.append(header)
    regex = re.compile("^" + regex + "$")
    return headers, regex


def log_to_dataframe(log_file, regex, headers):
    """
    Function to transform log file to dataframe.

    return: The log dataframe.
    """
    log_messages = []
    linecount = 0
    cnt = 0
    with open(log_file, "r", encoding="utf8", errors="ignore") as fin:
        for line in fin.readlines():
            cnt += 1
            try:
                match = regex.search(line.strip())
                message = [match.group(header) for header in headers]
                log_messages.append(message)
                linecount += 1
            except Exception as e:
                pass
    print("Total size after encoding is", linecount, cnt)
    logdf = pd.DataFrame(log_messages, columns=headers, dtype=str)
    return logdf


def load_data(filename, log_format):
    """
    Loads log from given file and format.

    :param filename: Files to read.
    :param log_format: Target log format.
    :return: The loaded log data.
    """
    headers, regex = generate_logformat_regex(log_format)
    df_log = log_to_dataframe(filename, regex, headers)
    return df_log



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from sklearn.preprocessing import LabelEncoder

from logai.algorithms.algo_interfaces import CategoricalEncodingAlgo


class LabelEncoding(CategoricalEncodingAlgo):
    """This is a wrapper class for LabelEncoder from scikit-learn library. For more details see 
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.LabelEncoder.html.
    """
    def __init__(self):
        """
        Initializes the label encoder.
        """
        self.model = LabelEncoder()

    def fit_transform(self, log_attributes: pd.DataFrame):
        """
        Fits and transforms log_attributes into label encoding categories.

        :param log_attributes: A list of log attributes in text format.
        :return: The label encoding categories.
        """

        res = pd.DataFrame()
        for feature_name in log_attributes.columns:
            x = self.model.fit_transform(log_attributes[feature_name])
            x_name = "{}_categorical".format(feature_name)
            res[x_name] = pd.Series(x, index=log_attributes[feature_name].index)
        return res



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
from attr import dataclass

import logai.algorithms.vectorization_algo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class VectorizerConfig(Config):
    """Config class for Vectorizer. 
    
    :param algo_name: The name of the vectorizer algorithm.
    :param algo_param: The parameters of the vectorizer algorithm .
    :param custom_param: Additional custom parameters to be passed to the vectorizer algorithm.
    """
    algo_name: str = "word2vec"
    algo_param: object = None
    custom_param: object = None

    @classmethod
    def from_dict(cls, config_dict):
        config = super(VectorizerConfig, cls).from_dict(config_dict)
        config.algo_param = factory.get_config(
            "vectorization", config.algo_name.lower(), config.algo_param
        )
        return config


class LogVectorizer:
    """
    Implement Log Vectorizer to transform raw log data to vectors. It Currently supports various statistical 
    (e.g. TfIdfVectorizer) and neural (Word2Vec, FastText, LogBERT) vectorizer models.
    """

    def __init__(self, config: VectorizerConfig):
        name = config.algo_name.lower()
        config_class = factory.get_config_class("vectorization", name)
        algorithm_class = factory.get_algorithm_class("vectorization", name)
        self.vectorizer = algorithm_class(
            config.algo_param if config.algo_param else config_class()
        )

    def fit(self, loglines: pd.Series):
        """Fit method for LogVectorizer, to train the vectorizer model on the training data.
        
        :param loglines: A pandas Series object containing the training raw log data.
        """
        self.vectorizer.fit(loglines)

    def transform(self, loglines: pd.Series) -> pd.Series:
        """Transform method for LogVectorizer, to transform the raw log text data to vectors.
        
        :param loglines: A pandas Series object containing the test raw log data.
        :return: A pandas Series object containing the vectorized log data.
        """
        return self.vectorizer.transform(loglines)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
from sklearn.metrics import accuracy_score, precision_score, recall_score


def get_accuracy_precision_recall(y: np.array, y_labels: np.array):
    """
    Evaluates the anomaly and labels.

    :param y: Model inference results.
    :param y_labels: y labels.
    :return: Accuracy, precision, and recall.
    """

    if len(y) != len(y_labels):
        raise IndexError("The length of anomalies and labels should be the same")

    accuracy = accuracy_score(y, y_labels)
    precision = precision_score(y, y_labels)
    recall = recall_score(y, y_labels)
    return accuracy, precision, recall
